//   GenericDiff.h -*- C++ -*- Stochastic simulation of QSD trajectories.
//     
//   Copyright (C) 2001  Todd Brun
//   
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//   
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//   
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//   ----------------------------------------------------------------------
//   If you improve the code or make additions to it, or if you have
//   comments or suggestions, please contact me:
//
//   Dr. Todd Brun			        Tel    +1-609-734-8335
//   Institute for Advanced Study		FAX    +1-609-951-4489
//   Einstein Drive				email  tbrun@ias.edu
//   Princeton, NJ  08540  USA
/////////////////////////////////////////////////////////////////////////////

#ifndef _Generic_hhh
#define _Generic_hhh 1

#include <stdio.h>
#include "Operator.h"
#include "CmplxRan.h"
#include "Traject.h"

typedef Operator (*AlphaFunc)(double, State&, Operator&, int, Operator*);
typedef Operator (*BetaFunc)(double, double, State&, Operator&, int,
  Operator*, ComplexRandom*);

class GenericDiff: public AdaptiveStep
{
public:
  GenericDiff( const State& psi, const Operator& theH, int theNL,
      const Operator* theL, double theEpsilon = 0.000001,
      AlphaFunc theAlpha = 0, BetaFunc theBeta = 0)
      : AdaptiveStep(psi,theH,theNL,theL,theEpsilon)
    { alpha=theAlpha;
      beta=theBeta;
    }
  ~GenericDiff();
  virtual void operator()(State& psi, double t, double dt,
      double& dtlast, ComplexRandom* rndm);
protected:
  AlphaFunc alpha;
  BetaFunc beta;
  virtual void derivs(double, State&, State&);
  virtual void rkck(double t, double h);
  virtual void rkqs(double& t, double htry, double eps,
	double& hdid, double& hnext);
  virtual void odeint(State& ystart, double t1, double t2, double eps,
	double& h1, double hmin, int& nok, int& nbad);
};
// This is the stepper for a generic diffusion equation
//
//    |dpsi> = alpha(psi)|psi>dt + beta(psi)|psi>dW
//
// where alpha and beta are operators which may depend on the state |psi>
// and the time; dW represents some number of real Gassian processes with
// mean dW^2 = dt.  The routines to calculate the operators alpha and beta
// are supplied by the user; they are assumed to be norm-preserving, which
// requires that
//
//   <alpha + alpha^dagger + beta^dagger beta> = 0
//   <beta + beta^dagger> = 0
//
// The arrays L and Ldag for this generic diffusion should not be interpreted
// as Lindblad operators, but merely as generators used in calculating
// alpha and beta.  The numerical integration works exactly the same
// as in the QSD equation. 

class GenericStochDiff: public GenericDiff
{
public:
  GenericStochDiff( const State& psi, const Operator& theH, int theNL,
      const Operator* theL, double theEpsilon = 0.000001,
      AlphaFunc theAlpha = 0, BetaFunc theBeta = 0)
      : GenericDiff(psi,theH,theNL,theL,theEpsilon,theAlpha,theBeta) { }
  ~GenericStochDiff();
  virtual void operator()(State& psi, double t, double dt,
      double& dtlast, ComplexRandom* rndm);
protected:
  virtual void odeint(State& ystart, double t1, double t2, double eps,
	double& h1, double hmin, int& nok, int& nbad, ComplexRandom* rndm);
};
// This is another stepper for a generic diffusion equation
//
//    |dpsi> = alpha(psi)|psi>dt + beta(psi)|psi>dW
//
// where alpha and beta are operators which may depend on the state |psi>
// and the time; dW represents some number of real Gassian processes with
// mean dW^2 = dt.  The routines to calculate the operators alpha and beta
// are supplied by the user; they are assumed to be norm-preserving, which
// requires that
//
//   <alpha + alpha^dagger + beta^dagger beta> = 0
//   <beta + beta^dagger> = 0
//
// The arrays L and Ldag for this generic diffusion should not be interpreted
// as Lindblad operators, but merely as generators used in calculating
// alpha and beta.  The numerical integration works exactly the same
// as in the QSD equation. 
//
// This stepper differs from GenericDiff in that the stochastic steps as
// well as the deterministic steps have adaptive sizes.  These steppers
// can also be adapted to simulate equations with colored (i.e., time-
// correlated) noise.


#endif
