//   Matrix.h -*- C++ -*- Matrix algebra in Hilbert space.
//     
//   Copyright (C) 2007  Todd Brun, Ruediger Schack, Bilal Shaw
//   
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//   
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//   
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//   ----------------------------------------------------------------------
//   If you improve the code or make additions to it, or if you have
//   comments or suggestions, please contact us:
//
//  
//
//   Dr. Todd Brun
//   Communication Sciences Institute           Tel    +001-213-740-3503
//   Department of Electrical Engineering       FAX    +001-213-740-8729 
//   University of Southern California          email  tbrun@usc.edu
//   3740 McClintock Ave, EEB 502
//   Los Angeles, CA, 90089-2565
//   USA			        
//
//   Bilal Shaw                                Tel    +001-213-740-5564
//   Communication Sciences Institute          email  bilalsha@usc.edu
//   Department of Electrical Engineering
//   University of Southern California
//   3740 McClintock Ave, EEB 514
//   Los Angeles, CA, 90089
//
/////////////////////////////////////////////////////////////////////////////


#ifndef MATRIX_H 
#define MATRIX_H

#include "Complex.h"
#include "State.h"
#include <string>

#ifndef NON_GNU_PROTOTYPE
  class PrimaryOperator;
  class Operator;
  class CPmap;
#else
  extern class PrimaryOperator;
  extern class Operator;
  extern class CPmap;   
#endif

//State and Matrix are friends of each other

//Matrix is a derived class of State. The data members of State are rendered
//protected so that Matrix can have access to them. 


class Matrix : public State{
// The Matrix class represents quantum states in the density operator form with a 
// particular choice of Hilbert space; this can include varying numbers of physical
// degrees of freedom, described varying numbers of basis states. 

public:

  Matrix();
  // Default constructor; produce a State of size 0x0
  
  Matrix(const Matrix& rho);
  // copy constructor
  
  Matrix(int n, FreedomType form=FIELD);
  // Produces a single degree-of-freedom Matrix with n basis states;
  // all basis states have amplitude 0 except the ground state,
  // which has amplitude 1; form gives the FreedomType of the Matrix
  // (default is FIELD)

  Matrix(const State& psi);
  // This creates a Matrix of the type |psi><psi|      
    
  Matrix(int n, int* dimensions, FreedomType* forms);
  // Multiple degree of freedom ground Matrix.  n gives the number of
  // degrees of freedom; dimensions is an array of n integers, which
  // specify the number of basis states to allocate for each degree
  // of freedom; forms is an array of n FreedomTypes, giving the
  // physical type of each degree of freedom.  The basis of the full
  // n-freedom state is given by the products of the basis states of
  // the n freedoms; all have amplitude 0 except for the ground state.
  
  Matrix(int n, Complex* elements, FreedomType form=FIELD);
  // Produces a one degree-of-freedom Matrix with n basis
  // states.  elements is an array of n Complex numbers, representing 
  // the amplitudes of the n basis states; form gives the FreedomType
  // of the State (default is FIELD).
  
  Matrix(int n, Matrix* matrixList); 
  // Produces an n degree-of-freedom Matrix.  matrixList is an array of
  // n one-freedom matrices.  The n-freedom density matrix will be produced in a
  // product matrix of the n matrices in matrixList.  The FreedomTypes and
  // dimensions of the different freedoms of the n-freedom matrix will
  // match those of the one-freedom matrices in matrixList.

  Matrix(int n, int nstate, FreedomType form=FIELD); // Fock Density Matrix
  // Produces a single degree-of-freedom Matrix with n basis states;
  // all basis states have amplitude 0 except state number nstate,
  // which has amplitude 1; form gives the FreedomType of the Matrix
  // (default is FIELD).

  Matrix(int n, Complex alpha, FreedomType form=FIELD); // Coherent Density Matrix
  // Produces a one degree-of-freedom Matrix with n basis
  // states in a coherent state given by the Complex number alpha.
  // The Matrix is represented in a Fock (number) state basis.
  // form gives the FreedomType of the Matrix (default is FIELD).
  
  Matrix(int n, int nstate, Complex alpha, FreedomType form=FIELD);
  // Excited coherent state.  Produces a single degree-of-freedom State
  // with n basis states.  Basis state number nstate has amplitude 1,
  // all others have amplitude 0, and the state is in the excited coherent
  // state or displaced Fock state basis, centered at alpha in phase space.
  // form gives the FreedomType of the State (default is FIELD).    
  
  Matrix(int n, Complex alpha, double width, double dx,
        FreedomType form=PARTICLE);//Gaussian Density Matrix
  // Produces a one degree-of-freedom Matrix with n basis
  // states in a Gaussian state given by the Complex number alpha
  // and the given width.  The Matrix is represented in a position state
  // basis, on a lattice with spacing dx.  If dx==0 the spacing is assumed
  // to be 1.  form gives the FreedomType of the Matrix (default is PARTICLE).

  void productMatrix(int n, Matrix* matrixList); // create a product density matrix  
  // create a product (tensored) matrix
  
  
  ~Matrix();
	
  virtual void fock(int n, int nstate); // create a fock state
  virtual void coherent(int n, Complex alpha);// create a coherent state
  virtual void gaussian(int n, Complex alpha, double width, double dx);//creates a gaussian density matrix
	
  Matrix& operator=(const Matrix&); //assignment operator
  Matrix& operator=(int);
  // zero assignment; enables one to type rho=0 to set all the entries
  //of a d x d matrix to 0. Gives an error for any int other than 0.
	
	Matrix& operator*=(CPmap&); //multiplication by a CPmap  
  Matrix& operator*=(const Matrix&);//multiplication by a matrix 
  Matrix& operator*=(const Complex&);//multiplication by a complex z*rho
  Matrix& operator*=(double);//multiplication by a real x*rho 							
  Matrix& operator*=(ImaginaryUnit);// i*rho
  Matrix& operator+=(const Matrix&);//add a matrix
  Matrix& operator-=(const Matrix&);//subtract a matrix									
  Complex Tr(); // Trace of a matrix
  Matrix& hc(); // Hermitian Conjugate of a matrix
  Matrix& transpose(); // Transpose of a matrix
  Matrix& conjugate(); // Conjuagate of a matrix
	Matrix reducedMatrix(int); // Returns the reduced Matrix specified for that dimension
  
  //Complex* operator[](int);
  
  // Friend arithmetic operations

  friend Matrix operator*(Matrix&, CPmap&);// rho*A[k]
  friend Matrix operator*(CPmap&, Matrix&);// A[k]*rho
  friend Matrix operator*(const Matrix&, const Matrix&); //A*B
  friend Matrix operator*(const Complex&, const Matrix&); // z*rho 
  friend Matrix operator*(const Matrix&, const Complex&); // rho*z
  friend Matrix operator*(double, Matrix&); //x*rho
  friend Matrix operator*(const Matrix&, double); // rho*x
  friend Matrix operator+(const Matrix&, const Matrix&);//A+B
  friend Matrix operator-(const Matrix&, const Matrix&);//A-B
  friend Matrix operator+(const Matrix&); // unary +A
  friend Matrix operator-(const Matrix&); // unary -A
	
// Friend I/O operations
     	
  friend ostream& operator<<(ostream&, const Matrix&);
  //outputs a Matrix in a standard ASCII form.  This can be used to 
  // save and recover results of a calculation
  friend istream& operator>>(istream&, Matrix&);
  // inputs a Matrix in a standard ASCII form.  This can be used to 
  // save and reover results of a calculation.

     			                                                  		
  int size();
  int getSize(int = 0);
  int getmySize();
  int getmaxSize();
  
  
  
  void printMatrix();
  
  Complex& getdata(int, int);
  void setdata(int, int, Complex);
  virtual void xerox(Matrix&);
  FreedomType getmyType();
  virtual void diagnostic();
  virtual void normalize();
  virtual void error(const char*) const;	// print out error message and exit
  
  
  // Basis changing functions
  
  virtual Complex centerCoords();
    // return center of coordinates
  virtual Complex getCoords(int = 0);
    // center of coordinates of nth freedom (default 0)
  virtual void setCoords(Complex&,int=0);
    // set the value of the coords for a freedom (default 0)
  virtual void displaceCoords(Complex&, int=0);
    // adds a Complex displacement to the center of
    // coordinates of a freedom (default freedom is 0)
  virtual double checkBounds(int, int=2);
    // check amplitudes of top basis states
    // (default: top 2 basis states)
  
  
   //virtual void moveCoords(const Complex& displacement, int theFreedom=0,
     // double shiftAccuracy=1e-4);
  
   virtual void moveCoords2(const Complex& displacement, int theFreedom=0);
  
   virtual void recenter(int theFreedom=0,double shiftAccuracy=1e-4);
  
   virtual void fieldRecenter(int theFreedom=0,double shiftAccuracy=1e-4);
  
   virtual void particleRecenter(int theFreedom=0,double shiftAccuracy=1e-4);
  
   virtual void moveToCoords(const Complex&, int theFreedom=0,
      double shiftAccuracy=1e-4);
  
   virtual void centerOn(Matrix&,double shiftAccuracy=1e-4);
  
   virtual void adjustCutoff(int theFreedom=0, double epsilon=1e-4, int padSize=2);
    // adjust amount of storage used by the State by adjusting the
    // cutoff for freedom number theFreedom.
   
   virtual void fullSize();	// makes size of state match physical size of storage

  


/*********************************************************************
                             PRIVATE
*********************************************************************/	
private:
		
  
  virtual void copy(const Matrix&);
  virtual void free();
  
  // apply a PrimaryOperator to a Matrix
  virtual void apply(PrimaryOperator&, int, int, FreedomType, double);
  
  virtual void moveStep(Complex&, Complex&);	// shift coords by infinitesimal step
  virtual void stretchFreedom(int,int);		// increase storage used by a freedom
  virtual void shrinkFreedom(int,int);		// compact storage used by a freedom

  // These are friend classes as they need access to various functions
  // and data memebers of Matrix
	friend class State;
  friend class Operator;
  friend class CPmap;
};

#endif
