// qwalk.cc

// Quantum random walk.  One PARTICLE (position) degree of freedom, n SPINs
// (the quantum "coins").  Implements model of Aharonov, Ambainis, Kempe, etc.,
// with variations in the number of coins, intended to probe the
// quantum-classical transition.

#include <math.h>
#include <stdio.h>
#include <iostream.h>
#include <fstream.h>

#include "ACG.h"
#include "Traject.h"
#include "State.h"
#include "Operator.h"
#include "ParticleOp.h"
#include "SpinOp.h"
#include "Complex.h"

#define TWOPI 6.2831853

int main()
{
// Basic operators

  IdentityOperator id;
  PositionOperator x(0);	// Operates on the zeroth degree of freedom.
  ShiftOperator shift(0);
  SigmaX sx1(1);		// Operates on the first "coin"
  SigmaY sy1(1);
  SigmaZ sz1(1);
  SigmaPlus sp1(1);
  SigmaX sx2(2);		// Operates on the second "coin"
  SigmaY sy2(2);
  SigmaZ sz2(2);
  SigmaPlus sp2(2);
  SigmaX sx3(3);		// Operates on the third "coin"
  SigmaY sy3(3);
  SigmaZ sz3(3);
  SigmaPlus sp3(3);
  SigmaX sx4(4);		// Operates on the fourth "coin"
  SigmaY sy4(4);
  SigmaZ sz4(4);
  SigmaPlus sp4(4);
  SigmaX sx5(5);		// Operates on the fifth "coin"
  SigmaY sy5(5);
  SigmaZ sz5(5);
  SigmaPlus sp5(5);

  Operator sm1 = sp1.hc();
  Operator sm2 = sp2.hc();
  Operator sm3 = sp3.hc();
  Operator sm4 = sp4.hc();
  Operator sm5 = sp5.hc();

// Derived operators

  Operator x2 = x*x;

  Operator proj11 = sp1*sm1;	// Projector onto |1> for coin 1
  Operator proj01 = sm1*sp1;	// Projector onto |0> for coin 1
  Operator flip1 = sqrt(0.5)*(sx1+sz1);	// Hadamard transform for coin flip
  Operator walk1 = shift*proj01 + shift.hc()*proj11; // Random walk

  Operator proj12 = sp2*sm2;	// Projector onto |1> for coin 2
  Operator proj02 = sm2*sp2;	// Projector onto |0> for coin 2
  Operator flip2 = sqrt(0.5)*(sx2+sz2);	// Hadamard transform for coin flip
  Operator walk2 = shift*proj02 + shift.hc()*proj12; // Random walk

  Operator proj13 = sp3*sm3;	// Projector onto |1> for coin 3
  Operator proj03 = sm3*sp3;	// Projector onto |0> for coin 3
  Operator flip3 = sqrt(0.5)*(sx3+sz3);	// Hadamard transform for coin flip
  Operator walk3 = shift*proj03 + shift.hc()*proj13; // Random walk

  Operator proj14 = sp4*sm4;	// Projector onto |1> for coin 4
  Operator proj04 = sm4*sp4;	// Projector onto |0> for coin 4
  Operator flip4 = sqrt(0.5)*(sx4+sz4);	// Hadamard transform for coin flip
  Operator walk4 = shift*proj04 + shift.hc()*proj14; // Random walk

  Operator proj15 = sp5*sm5;	// Projector onto |1> for coin 5
  Operator proj05 = sm5*sp5;	// Projector onto |0> for coin 5
  Operator flip5 = sqrt(0.5)*(sx5+sz5);	// Hadamard transform for coin flip
  Operator walk5 = shift*proj05 + shift.hc()*proj15; // Random walk

// Arrays of evolution operator

  Operator evolve[5] = { walk1*flip1, walk2*flip2, walk3*flip3,
                         walk4*flip4, walk5*flip5 };

// Array of CNOT operators; cnot[i] performs a controlled-not from
// coin i onto coin i+1

  Operator cnot[5] = { id, proj01 + proj11*sx2,
                           proj02 + proj12*sx3,
                           proj03 + proj13*sx4,
                           proj04 + proj14*sx5 };

// Loop over different numbers of "coins"

  int numSteps = 500;			// Number of iterations of map
  int spaceSize = 1010;			// Size of lattice
  double q,q2,logquant;
  int theCoin;

  for(int nCoins=1; nCoins<=5; nCoins++) {

// The initial state

    State psi1(spaceSize,PARTICLE);	// 1-freedom state ("|x = 0>")
    State psi2(2,SPIN);			// 1-freedom state ("down")
    State psilist[6] = {psi1,psi2,psi2,psi2,psi2,psi2};       
    State psi(6,psilist);                 // Product state  ("|0>(|0>^5)")
    theCoin = 0;
    //psi *= sqrt(0.5)*(id+IM*sx1);	// Put the coins in a "cat" state
    //for( int iter=0; iter<nCoins; iter++)
      //psi *= cnot[iter];

    //psi *= sqrt(0.5)*(id+IM*sx1);	// Put the coins in a y eigenstate
    //psi *= sqrt(0.5)*(id+IM*sx2);
    //psi *= sqrt(0.5)*(id+IM*sx3);
    //psi *= sqrt(0.5)*(id+IM*sx4);
    //psi *= sqrt(0.5)*(id+IM*sx5);

// Loop over the evolution steps

    for( int count=1; count<=numSteps; count++ ) { // iterate over steps

      psi *= evolve[theCoin];	// flip the appropriate coin and evolve
      psi.normalize();
      q = real(psi*(x*psi));		// q = <x>
      q2 = real(psi*(x2*psi));		// q2 = <x^2>
      //if(count>1)
        //logquant = log(q2-q*q)/log(count);
      //else
        //logquant = 0;
      cout << count << "  " << q << "  " << q2 - q*q << endl;
        //<< "  " << logquant << endl;
      // cout << log(q2-q*q)/log(count+1) << "," << endl;
      theCoin++;		// switch to the next coin
      theCoin = theCoin%nCoins;
      psi.adjustCutoff(0,0.000001,5);	// use adjustable state cutoff

    } // end step loop

    cout << endl;
    psi.fullSize();
    int i,j;
    double xValue;
    for(i=numSteps; i>0; i-=2) {
      xValue = 0.0;
      for(j=0; j<32; j++) xValue += norm(psi[2*i-1+j*spaceSize]);
      cout << -i << "  " << xValue << endl;
    }
    for(i=0; i<=numSteps; i+=2) {
      xValue = 0.0;
      for(j=0; j<32; j++) xValue += norm(psi[2*i+j*spaceSize]);
      cout << i << "  " << xValue << endl;
    }
    cout << endl;

  }  // end coins loop

}

