// qwalk2.cc

// Quantum random walk.  One PARTICLE (position) degree of freedom, one SPIN
// (the quantum "coin").  Implements model of Aharonov, Ambainis, Kempe, etc.,
// with variations intended to probe the quantum-classical transition.

// In this particular variation, we simulate decoherence in the coin after
// each unitary "flip."  In the case of perfect decoherence we should recover
// the classical result; in the case of imperfect decoherence, we should
// interpolate between the quantum and classical cases.

// This decoherence is simulated by randomly choosing one of the following
// two unitary operators after each flip:
//   U1 = cos(theta) id + i sin(theta) sigma_z
//   U2 = cos(theta) id - i sin(theta) sigma_z .
// Averaging over many runs of this is the same as doing the weak measurement
//   A1 = cos(phi) |0><0| + sin(phi) |1><1|
//   A2 = sin(phi) |0><0| + cos(phi) |1><1|
// where
//   cos(theta) = (sin(phi) + cos(phi))/sqrt(2)
//   sin(theta) = (sin(phi) - cos(phi))/sqrt(2) .
// It is also equivalent to the weak measurement
//   B0 = sqrt(p) |0><0|
//   B1 = sqrt(p) |1><1|
//   B2 = sqrt(1-p) id
// with p = (sin(phi)-cos(phi))^2 = 2 sin(theta)^2 .

// Zero decoherence is given by theta=0, maximal decoherence by theta=pi/4 .

#include <math.h>
#include <stdio.h>
#include <iostream.h>
#include <fstream.h>

#include "ACG.h"
#include "Traject.h"
#include "State.h"
#include "Operator.h"
#include "ParticleOp.h"
#include "SpinOp.h"
#include "Complex.h"

#define TWOPI 6.2831853


double xFunc(double p)
{
  return (1.0-p)*(1.0-p)/(p*(2.0-p));
}

double x2Func1(double p)
{
  return 1.0 + 2.0*(1.0-p)*(1.0-p)/((2.0-p)*p);
}

double x2Func2(double p)
{
  return (1.0-p)*(1.0-p)*(7.0 + (1.0-p)*(1.0-p))/(p*p*(2.0-p)*(2.0-p));
}

int main()
{
// Basic operators

  IdentityOperator id;
  PositionOperator x(0);	// Operates on the zeroth degree of freedom.
  ShiftOperator shift(0);
  SigmaX sx(1);			// Operates on the "coin"
  SigmaY sy(1);
  SigmaZ sz(1);
  SigmaPlus sp(1);

  Operator sm = sp.hc();

// Derived operators

  Operator x2 = x*x;

  Operator proj0 = sm*sp;	// Projector onto |0> for coin 1
  Operator proj1 = sp*sm;	// Projector onto |1> for coin 1
  Operator flip = sqrt(0.5)*(sx+sz);	// Hadamard transform for coin flip
  Operator walk = shift*proj0 + shift.hc()*proj1; // Random walk

  Operator evolve = walk*flip;		// the full evolution operator

// Necessary constants and data structures

  int i,j;
  const int numSteps = 500;		// Number of iterations of map
  const int spaceSize = 1010;		// Size of lattice
  const int numLevels = 4;		// Number of decoherence levels
  const int numRuns = 10000;		// Number of runs per level
  //double thetaStep = M_PI/16;		// decoherence interval
  double thetaStep = M_PI/64;		// decoherence interval
  double q[numSteps];			// average M[<q>] for each step
  double q2[numSteps];			// average M[<q^2>] for each step
  double xValue[2*numSteps+1];		// average M[|psi(x)|^2] at t=500
  double p,theta;
  double xConst,x2Const1,x2Const2;

  ACG gen(837234);			// Random Number Generator

// Loop over different levels of decoherence

  for(int nLevel=1; nLevel<=numLevels; nLevel++) {

// Define the randomly-chosen unitaries

    theta = nLevel*thetaStep;
    p = 2.0*sin(theta)*sin(theta);
    xConst = xFunc(p);
    x2Const1 = x2Func1(p);
    x2Const2 = x2Func2(p);
    Operator unitary[2] = { cos(theta)*id + IM*(sin(theta)*sz),
                            cos(theta)*id - IM*(sin(theta)*sz) };
    for(i=0; i<numSteps; i++) {
      q[i] = 0.0;
      q2[i] = 0.0;
      xValue[2*i] = 0.0;
      xValue[2*i+1] = 0.0;
    }
    xValue[2*numSteps] = 0.0;

// Loop over the number of runs

    for(int nRun=0; nRun<numRuns; nRun++) {

// The initial state

      State psi1(spaceSize,PARTICLE);	// 1-freedom state ("|x = 0>")
      State psi2(2,SPIN);		// 1-freedom state ("down")
      State psilist[2] = {psi1,psi2};       
      State psi(2,psilist);             // Product state  ("|0>|0>")
      //psi *= sqrt(0.5)*(id + IM*sx);  // starting y eigenstate

// Loop over the evolution steps

      for( int count=0; count<numSteps; count++ ) { // iterate over steps

        psi *= evolve;			// flip the coin and walk
        if(gen.asDouble() > 0.5)
          psi *= unitary[0];
        else
          psi *= unitary[1];
        psi.normalize();
        q[count] += real(psi*(x*psi));		// q = M[<x>]
        q2[count] += real(psi*(x2*psi));	// q2 = M[<x^2>]

      } // end step loop

      for(i=0; i<=2*numSteps; i++)
        for(j=0; j<2; j++) xValue[i] += norm(psi[i+j*spaceSize]);

    }  // end runs loop

    for(i=numSteps; i>0; i-=2)
      cout << -i << "  " << xValue[2*i-1]/numRuns << endl;
    for(i=0; i<=numSteps; i+=2)
      cout << i << "  " << xValue[2*i]/numRuns << endl;
    cout << endl;

    for(i=0; i<numSteps; i++) {
      q[i] /= numRuns;
      q2[i] /= numRuns;
      cout << (i+1) << "  " << q[i] << "  " << q2[i] - q[i]*q[i]
        << "  " << xConst << " " << x2Const1*(i+1)-x2Const2 << endl;
    }
    cout << endl;

  } // end decoherence levels loop

} // end program

