// qwalk2.cc

// Quantum random walk.  One PARTICLE (position) degree of freedom, one SPIN
// (the quantum "coin").  Implements model of Aharonov, Ambainis, Kempe, etc.,
// with variations intended to probe the quantum-classical transition.

// In this particular variation, we simulate decoherence in the coin after
// each unitary "flip."  In the case of perfect decoherence we should recover
// the classical result; in the case of imperfect decoherence, we should
// interpolate between the quantum and classical cases.

// The particular model of decoherence we use has a small probability p
// of measuring the coin at each step; after the coin is measured, it is
// reset to the starting state |0>.
// Averaging over many runs of this is the same as doing the weak measurement

// Zero decoherence is given by p=0, maximal decoherence by p=1.

#include <math.h>
#include <stdio.h>
#include <iostream.h>
#include <fstream.h>

#include "ACG.h"
#include "Traject.h"
#include "State.h"
#include "Operator.h"
#include "ParticleOp.h"
#include "SpinOp.h"
#include "Complex.h"

#define TWOPI 6.2831853

const double cVal = 1.0 - sqrt(0.5);
const double coeff = 2*cVal*(1.0-cVal);
const double c0 = 1.0;
const double c1 = 2.0/3.0;
const double c2 = 0.25;
const double c3 = 1.0/15.0;

double analytical(double p, int t)
{
  double temp1,temp2;
  temp1 = p*t/(1.0-p);
  temp2 = (1.0-1.0/temp1)/temp1;
  temp2 *= exp(temp1);
  temp2 += 1.0/(temp1*temp1);
  //temp2 = c3;
  //temp2 *= temp1;
  //temp2 += c2;
  //temp2 *= temp1;
  //temp2 += c1;
  //temp2 *= temp1;
  //temp2 += c0;
  //return coeff*temp2*t*t*pow(1.0-p,t);
  return coeff*t*t*temp2*pow(1.0-p,t);
}

int main()
{
// Basic operators

  IdentityOperator id;
  PositionOperator x(0);	// Operates on the zeroth degree of freedom.
  ShiftOperator shift(0);
  SigmaX sx(1);			// Operates on the "coin"
  SigmaY sy(1);
  SigmaZ sz(1);
  SigmaPlus sp(1);

  Operator sm = sp.hc();

// Derived operators

  Operator x2 = x*x;

  Operator proj0 = sm*sp;	// Projector onto |0> for coin 1
  Operator proj1 = sp*sm;	// Projector onto |1> for coin 1
  Operator flip = sqrt(0.5)*(sx+sz);	// Hadamard transform for coin flip
  Operator walk = shift*proj0 + shift.hc()*proj1; // Random walk

  Operator evolve = walk*flip;		// the full evolution operator

  Operator reset0 = proj0;		// Operators following a measurement
  Operator reset1 = sm*proj1;

// Necessary constants and data structures

  int i,j;
  const int numSteps = 500;		// Number of iterations of map
  int spaceSize = 1010;			// Size of lattice
  int numLevels = 5;			// Number of decoherence levels
  int numRuns = 200;			// Number of runs per level
  double pStep = 0.01;			// decoherence interval
  double q[numSteps];			// average M[<q>] for each step
  double q2[numSteps];			// average M[<q^2>] for each step
  //double xValue[2*numSteps+1];		// average M[|psi(x)|^2] at t=500
  double p,prob0;

  //ACG gen(837234);			// Random Number Generator
  ACG gen(23853);			// Random Number Generator

// Loop over different levels of decoherence

  for(int nLevel=1; nLevel<=numLevels; nLevel++) {

    p = nLevel*pStep;
    for(i=0; i<numSteps; i++) {
      q[i] = 0.0;
      q2[i] = 0.0;
      //xValue[2*i] = 0.0;
      //xValue[2*i+1] = 0.0;
    }
    //xValue[2*numSteps] = 0.0;

// Loop over the number of runs

    for(int nRun=0; nRun<numRuns; nRun++) {

// The initial state

      State psi1(spaceSize,PARTICLE);	// 1-freedom state ("|x = 0>")
      State psi2(2,SPIN);			// 1-freedom state ("down")
      State psilist[2] = {psi1,psi2};       
      State psi(2,psilist);                 // Product state  ("|0>|0>")

// Loop over the evolution steps

      for( int count=0; count<numSteps; count++ ) { // iterate over steps

        psi *= evolve;		// flip the coin and walk
        if(gen.asDouble() < p) {	// Decohere?
          prob0 = real(psi*(proj0*psi));
          if(gen.asDouble() < prob0)
            psi *= reset0;
          else
            psi *= reset1;
        }
        psi.normalize();
        q[count] += real(psi*(x*psi));		// q = M[<x>]
        q2[count] += real(psi*(x2*psi));	// q2 = M[<x^2>]

      } // end step loop

      //for(i=0; i<=2*numSteps; i++)
        //for(j=0; j<2; j++) xValue[i] += norm(psi[i+j*spaceSize]);

    }  // end runs loop

    //for(i=numSteps; i>0; i--)
      //cout << -i << "  " << xValue[2*i-1]/numRuns << endl;
    //for(i=0; i<=numSteps; i++)
      //cout << i << "  " << xValue[2*i]/numRuns << endl;
    //cout << endl;

    for(i=0; i<numSteps; i++) {
      q[i] /= numRuns;
      q2[i] /= numRuns;
      cout << i+1 << "  " << q[i] << "  " << q2[i] - q[i]*q[i]
        << "  " << analytical(p,i+1) << endl;
    }
    cout << endl;

  } // end decoherence levels loop

} // end program

