// qwalk3.cc

// Quantum random walk.  One PARTICLE (position) degree of freedom, n SPINs
// (the quantum "coins").  Implements model of Aharonov, Ambainis, Kempe, etc.,
// with variations in the number of coins, intended to probe the
// quantum-classical transition.  This version utilizes a large number
// of coins.

#include <math.h>
#include <stdio.h>
#include <iostream.h>
#include <fstream.h>

#include "ACG.h"
#include "Traject.h"
#include "State.h"
#include "Operator.h"
#include "ParticleOp.h"
#include "SpinOp.h"
#include "Complex.h"

#define TWOPI 6.2831853

#include "choose.cc"

int main()
{
// Basic operators

  IdentityOperator id;
  PositionOperator x(0);	// Operates on the zeroth degree of freedom.
  ShiftOperator shift(0);
  SigmaX sx(1);			// Operates on the "coin"
  SigmaY sy(1);
  SigmaZ sz(1);
  SigmaPlus sp(1);

  Operator sm = sp.hc();

// Derived operators

  Operator x2 = x*x;

  Operator proj1 = sp*sm;	// Projector onto |1> for coin
  Operator proj0 = sm*sp;	// Projector onto |0> for coin
  Operator flip = sqrt(0.5)*(sx+sz);	// Hadamard transform for coin flip
  Operator walk = shift*proj0 + shift.hc()*proj1; // Random walk

// Evolution operators

  Operator evolve = walk*flip;
  Operator reset0 = proj0;
  Operator reset1 = sm*proj1;
  //Operator reset0 = sqrt(0.5)*(id + IM*sx)*proj0;
  //Operator reset1 = sqrt(0.5)*(sx + IM*id)*proj1;
  const int nPowers = 101;
  Operator evolve0[nPowers];
  Operator evolve1[nPowers];
  evolve0[0] = id;
  evolve1[0] = id;
  int i;
  for(i=1; i<nPowers; i++) {
    evolve0[i] = reset0*evolve.pow(i);
    evolve1[i] = reset1*evolve.pow(i);
  }

// The initial state

  const int spaceSize = 510;		// Size of lattice
  State psi1(spaceSize,PARTICLE);	// 1-freedom state ("|x = 0>")
  State psi2(2,SPIN);			// 1-freedom state ("down")
  State psilist[2] = {psi1,psi2};
  State psiIni(2,psilist);		// Product state  ("|0>(|0>^5)")
  //psiIni *= sqrt(0.5)*(id+IM*sx);	// Put the coin in a y eigenstate

//  Output probability distribution

  double p[spaceSize];

// Loop over different numbers of "coins"

  //int numSteps = 64;			// Number of iterations of map
  int numSteps = 500;			// Number of iterations of map
  int tDivM, tModM;
  int j,k;
  double combinatorics;

  //for(int nCoins=1; nCoins<=5; nCoins++) {
  //for(int nCoins=1; nCoins<=64; nCoins *= 2) {
  for(int nCoins=numSteps/2; nCoins<=numSteps/2; nCoins++) {

// Loop over the evolution steps

    cout << "0  0  0" << endl;
    for( int t=nCoins; t<=numSteps; t += nCoins ) { // iterate over steps

      tDivM = t/nCoins;
      tModM = t%nCoins;
      double xBar=0.0;
      double x2Bar=0.0;
      for( k=0; k<spaceSize; k++) p[k] = 0.0;
      for( i=0; i<=nCoins; i++ ) {		// Coins flipped tDivM times
        State psi = psiIni;
        for( k=0; k<i; k++)  psi *= evolve0[tDivM];
        for( k=0; k<nCoins-i; k++)  psi *= evolve1[tDivM];
        combinatorics = bico(nCoins,i);
        xBar += combinatorics*real(psi*(x*psi));	// <x>
        x2Bar += combinatorics*real(psi*(x2*psi));	// <x^2>
        for( k=0; k<spaceSize; k++) {			// p(x,t)
          p[k] += combinatorics*norm(psi[k]);
          p[k] += combinatorics*norm(psi[k+spaceSize]);
        }
      }
      cout << t << "  " << xBar << "  " << x2Bar-xBar*xBar << endl;
    } // end step loop
    cout << endl;
    for(i=numSteps; i>0; i--)
      cout << -i << "  " << p[i+i-1] << endl;
    for(i=0; i<=numSteps; i++)
      cout << i << "  " << p[i+i] << endl;
    cout << endl;
  }  // end coins loop
}  // end program

