//   CPmap.cc -*- C++ -*- Completely Positive Maps in Hilbert space. 
//     
//   Copyright (C) 2007  Todd Brun, Bilal Shaw
//   
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//   
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//   
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//   ----------------------------------------------------------------------
//   If you improve the code or make additions to it, or if you have
//   comments or suggestions, please contact us:
//
//  
//
//   Dr. Todd Brun
//   Communication Sciences Institute           Tel    +001-213-740-3503
//   Department of Electrical Engineering       FAX    +001-213-740-8729 
//   University of Southern California          email  tbrun@usc.edu
//   3740 McClintock Ave, EEB 502
//   Los Angeles, CA, 90089-2565
//   USA			        
//
//   Bilal Shaw                                Tel    +001-213-740-5564
//   Communication Sciences Institute          email  bilalsha@usc.edu
//   Department of Electrical Engineering
//   University of Southern California
//   3740 McClintock Ave, EEB 514
//   Los Angeles, CA, 90089
/////////////////////////////////////////////////////////////////////////////


#include <math.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <string>
#include "CPmap.h"
#include "Matrix.h"
#include "PrimOp.h"
#include "Operator.h"

using namespace std;

//Default constructor
CPmap::CPmap()
{
  cpmapSize = 0;
  cpmapPointer = 0;
}

CPmap::CPmap(int n, Operator* oparray)
{
  cpmapSize = n;
  cpmapPointer = new Operator[cpmapSize]; // allocate space for array of Kraus operators
  
  for(int k = 0; k < cpmapSize; k++)
  {
    cpmapPointer[k] = oparray[k];
  } 
}

//copy constructor
CPmap::CPmap(const CPmap& A)
{
  
  cpmapSize = 0;
  copy(A);
  
}


CPmap::~CPmap()
{
  
  free();
  
}

void CPmap::free()
{
  if(cpmapPointer != 0)
  {
#ifndef NON_GNU_DELETE
    delete[] cpmapPointer;
#else // ---- NON GNU CODE ----
    delete [cpmapSize] cpmapPointer;		// NON GNU recycle storage
#endif
  
  cpmapPointer = 0;
  cpmapSize = 0;
  }
}

    
CPmap& CPmap::operator=(const CPmap& A) //assignment operator
{
  
  if (this != &A)// otherwise cpmap = cpmap would lead to trouble
  {
    if((cpmapSize != A.cpmapSize))
    {
      free();//recycle any memory
    }
		
   copy(A);//copy rho into b
  }
	
  return *this;
}

void CPmap::copy(const CPmap& A)
{
  if((cpmapSize != A.cpmapSize))
  {
    cpmapSize = A.cpmapSize;//if no memory is allocated, allocate it
    cpmapPointer = new Operator[cpmapSize];
  }
  
  for(int k = 0; k < cpmapSize; k++)
  {
    cpmapPointer[k] = A.cpmapPointer[k];
  }
}
    
CPmap operator*(const Complex& z, const CPmap& A)
{
  
  CPmap result(A);
  return (result *= z);
}    

CPmap operator*(const CPmap& A, const Complex& z)
{
  
  CPmap result(A);
  return (result *= z);
}

CPmap operator*(CPmap& A, double x)
{
  CPmap result(A);
  return (result *= x);
}

CPmap operator*(double x, CPmap& A)
{
  CPmap result(A);
  return (result *= x);
}

CPmap operator*(const CPmap& A, CPmap& B)
{
  CPmap result(A);
  return(result *= B);
}
   
// adding two CPmaps
// (A+B)*rho = A*rho + B*rho  
CPmap operator+(const CPmap& A, CPmap& B)
{
  CPmap result(A);
  return (result += B);
}

// (A-B)*rho = A*rho - B*rho
CPmap operator-(const CPmap& A, CPmap& B)
{
  CPmap result(A);
  return (result -= B);
}

// + unary operator
CPmap operator+(const CPmap& A)
{
  CPmap result(A);
  return result;
}

// - unary operator
CPmap operator-(const CPmap& A)
{
  CPmap result(A);
  int size = result.getcpmapSize();
  for(int j = 0; j < size; j++)
  {
    result.cpmapPointer[j] = -result.cpmapPointer[j];
  }
  
  return result;
}

// product of a CPmap and a double
CPmap& CPmap::operator*=(double x)  
{
  
  for(int j = 0; j < cpmapSize; j++)
  {
    cpmapPointer[j] *= x;
  }
  
  return *this;
}

// product of CPmap and a complex number
CPmap& CPmap::operator*=(const Complex& z)  
{
  
  for(int j = 0; j < cpmapSize; j++)
  {
    cpmapPointer[j] *= z;
  }
  
  return *this;
}

// multiplying two CPmaps
CPmap& CPmap::operator*=(CPmap& A)
{
  
  int sizeA = A.getcpmapSize();
  int sizethis = this->getcpmapSize();
  int productofsizes = sizeA*sizethis;
  Operator* Aoparrayptr = A.getcpmapPointer();
  Operator* oparrayptr = NULL;
  oparrayptr = new Operator[productofsizes];
  
  //CPmap* mymap = new CPmap(productofsizes, oparrayptr);
  
  int k = 0;
  
  for(int i = 0; i < sizethis; i++){
    for(int j = 0; j < sizeA; j++){
      oparrayptr[k] = Aoparrayptr[j]*cpmapPointer[i];
      k++;
    }
  }
  
  cpmapPointer = oparrayptr;
  cpmapSize = productofsizes;
  
  //delete  
  return *this;
}

// adding two CPmaps
// (A+B)*rho = A*rho + B*rho
CPmap& CPmap::operator+=(CPmap& B)
{
  
  int cpmapSizeA = this->getcpmapSize();
  int cpmapSizeB = B.getcpmapSize();
  int newSize = cpmapSizeA + cpmapSizeB;
  Operator* Boparrayptr = B.getcpmapPointer();
  
  // create a new array of operators
  Operator* newArray = new Operator[newSize];
  
  for (int i = 0; i < cpmapSizeA; i++)
  {
    newArray[i] = cpmapPointer[i];
  }
  
  int k = 0;
  for(int j = cpmapSizeA; j < newSize; j++)
  {
    
    newArray[j] = Boparrayptr[k];
    k++;
  }
    
  this->cpmapPointer = newArray;
  this->setcpmapSize(newSize);
  return *this;
}

//subtracting two CPmaps
// (A-B)*rho = A*rho - B*rho
CPmap& CPmap::operator-=(CPmap& B)
{
  int cpmapSizeA = this->getcpmapSize();
  int cpmapSizeB = B.getcpmapSize();
  int newSize = cpmapSizeA + cpmapSizeB;
  Operator* Boparrayptr = B.getcpmapPointer();
  
  // create a new array of operators
  Operator* newArray = new Operator[newSize];
  
  for (int i = 0; i < cpmapSizeA; i++)
  {
    newArray[i] = cpmapPointer[i];
  }
  
  int k = 0;
  for(int j = cpmapSizeA; j < newSize; j++)
  {
    
    newArray[j] = -Boparrayptr[k];
    k++;
  }
    
  this->cpmapPointer = newArray;
  this->setcpmapSize(newSize);
  return *this;
}
