//   GenericDiff.cc -- Stochastic simulation of QSD trajectories.
//     
//   Copyright (C) 2001  Todd Brun
//   
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//   
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//   
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//   ----------------------------------------------------------------------
//   If you improve the code or make additions to it, or if you have
//   comments or suggestions, please contact us:
//
//   Dr. Todd Brun			        Tel    +1-609-734-8335
//   Institute for Advanced Study               FAX    +1-609-951-4489
//   Einstein Drive                             email  tbrun@ias.edu
//   Princeton, NJ  08540  USA
/////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <iostream.h>
#include <fstream.h>

#include "Complex.h"
#include "State.h"
#include "Operator.h"
#include "ACG.h"
#include "Normal.h"
#include "Traject.h"
#include "GenericDiff.h"

#define MAXSTP 10000
#define TINY 1.0e-30
#define SAFETY 0.9
#define PGROW -0.2
#define PSHRNK -0.25
#define ERRCON 1.89e-4

GenericDiff::~GenericDiff()
{
  alpha = 0;
  beta = 0;
}

void GenericDiff::operator()(State& psi, double t, double dt,
	double& dtlast, ComplexRandom* rndm)
//
// This member function calculates a single QSD integration step.  It
// invokes the member function odeint (adapted from the Numerical Recipes
// routine) to advance the deterministic part of the equation in time,
// and uses the supplied function rndm to advance the stochastic
// part (with noise dxi).  The state psi is evolved from t to t+dt; dtlast
// contains the actual physical stepsize used by the adaptive stepsize
// routine.  (In case of a constant stepsize stepper, dtlast=dt.)
//
{
  int nok, nbad;

  double sqrtdt = sqrt(dt);
  if( rndm != 0 )				// any noise?
    stochasticFlag = 1;
	// set flag for derivs to generate stoch. part and store in newsum
  newsum = psi;					// initialize temp
  newsum = 0;
  odeint(psi,t,t+dt,epsilon,dtlast,0.0,nok,nbad,rndm);
	// generate deterministic evolution (and also stochastic term)
  newsum *= sqrtdt;
  psi += newsum;				// add stochastic term
  psi.normalize();
}

void GenericDiff::rkck(double t, double h, ComplexRandom* rndm)
//
// Given values for a state y and its derivative dydt known at t, use the
// fifth-order Cash-Karp Runge-Kutta method to advance the solution over an
// interval h and return the incremented state as yout.  Also return an
// estimate of the local truncation error in yout using the embedded
// fourth-order method.  The user supplies the routine derivs(t,y,dydt,rndm)
// which returns the derivative dydt at t.
//
// Adapted from the Numerical Recipes routine RKCK.C
//
{
  static double a2=0.2,a3=0.3,a4=0.6,a5=1.0,a6=0.875,b21=0.2,
	b31=3.0/8.0,b32=9.0/40.0,b41=4.0,b42 = -4.0,b43=1.2,
	b51 = -55.0/81.0, b52=-25.0/9.0, b53 = -175.0/81.0,b54=35.0/27.0,
	b61=-1631.0/11264.0,b62=35.0/256.0,b63=-115.0/7168.0,
	b64=1265.0/4096.0,b65=253.0/4096.0,c1=37888.0/11417.0,
	c3=5120.0/529.0,c4=10240.0/19481.0,c6=512.0/1771.0,
	dc5 = -554.0/1771.0;
  double dc1=-831.0/18944.0,dc3=831.0/17920.0,
	dc4=-831.0/5120.0,dc6=277.0/2048.0;

  ytemp = y;
  derivs(t,ytemp,dydt,rndm);
  dydt *= (h*b21);
  ytemp += dydt;				// First step
  derivs(t+a2*h,ytemp,ak2,rndm);		// Second step
  ytemp = y;
  dydt *= b31;
  ak2 *= (h*b32);
  ytemp += dydt;
  ytemp += ak2;
  derivs(t+a3*h,ytemp,ak3,rndm);		// Third step
  dydt *= b41;
  ak2 *= b42;
  ak3 *= (h*b43);
  ytemp = y;
  ytemp += dydt;
  ytemp += ak2;
  ytemp += ak3;
  derivs(t+a4*h,ytemp,ak4,rndm);		// Fourth step
  dydt *= b51;
  ak2 *= b52;
  ak3 *= b53;
  ak4 *= (h*b54);
  ytemp = y;
  ytemp += dydt;
  ytemp += ak2;
  ytemp += ak3;
  ytemp += ak4;
  derivs(t+a5*h,ytemp,ak5,rndm);		// Fifth step
  dydt *= b61;
  ak2 *= b62;
  ak3 *= b63;
  ak4 *= b64;
  ak5 *= (h*b65);
  ytemp = y;
  ytemp += dydt;
  ytemp += ak2;
  ytemp += ak3;
  ytemp += ak4;
  ytemp += ak5;
  derivs(t+a6*h,ytemp,ak2,rndm);		// Sixth step
//
// Accumulate increments with proper weights
//
  dydt *= c1;
  ak3 *= c3;
  ak4 *= c4;
  ak2 *= (h*c6);
  yout = y;
  yout += dydt;
  yout += ak3;
  yout += ak4;
  yout += ak2;
//
// Estimate error as difference between fourth and fifth order methods.
//
  dydt *= dc1;
  ak3 *= dc3;
  ak4 *= dc4;
  ak5 *= dc5;
  ak2 *= dc6;
  yerr = dydt;
  yerr += ak3;
  yerr += ak4;
  yerr += ak5;
  yerr += ak2;
}

void GenericDiff::rkqs(double& t, double htry, double eps,
	double& hdid, double& hnext, ComplexRandom* rndm)
//
// Fifth-order Runge-Kutta step with monitoring of local truncation error to
// ensure accuracy and adjust stepsize.  Input are the State y and its
// derivative dydt at the starting value of the time t.  Also input are the
// stepsize to be attempted, htry, and the required accuracy, eps. On output,
// y and t are replaced by their new values, hdid is the stepsize that was
// actually accomplished, and hnext is the estimated next stepsize.  derivs
// is the user-supplied routine that computes the right-hand side derivative.
//
// Adapted from the Numerical Recipes routine RKQS.C
//
{
  double errmax,h,tnew;

  h=htry;
  for (;;) {
	rkck(t,h,rndm);
	errmax = sqrt(real(yerr*yerr));
	errmax /= eps;
	if (errmax > 1.0) {
		h=SAFETY*h*pow(errmax,PSHRNK);
//		if (h < 0.1*h) h *= 0.1;
		tnew=t+h;
		if (tnew == t) {
		  cerr << "stepsize underflow in rkqs" << endl;
		  cerr << "errmax = " << errmax << endl;
		  cerr << "h = " << h << endl;
		  cerr << "eps = " << eps << endl;
		  exit(1);
		}
		continue;
	} else {
		if (errmax > ERRCON) hnext=SAFETY*h*pow(errmax,PGROW);
		else hnext=5.0*h;
		t += (hdid=h);
		y = yout;
		numdtsUsed++;
		break;
	}
  }
}

void GenericDiff::odeint(State& ystart, double t1, double t2, double eps, double& h1,
	double hmin, int& nok, int& nbad, ComplexRandom* rndm)
//
// Runge-Kutta driver with adaptive stepsize control  Integrate starting values
// ystart from t1 to t2 with accuracy eps.  h1 should be set as a guessed
// first stepsize, hmin as the minimum allowed stepsize (can be zero).  On
// output nok and nbad are the number of good and bad (but retried and fixed)
// steps taken, and ystart is replaced by values at the end of the integration
// interval.  derivs is the routine for calculating the right-hand side of
// the QSD equation, while rkqs is the name of the stepper routine
// to be used.
//
// Adapted from Numerical Recipes routine ODEINT.C
//
{
  int nstp;
  double t,hnext,hdid,h;

  t = t1;
  h = h1; if( h > (t2-t1) ) h = (t2-t1);
  nok = nbad = 0;
  y = ystart;
  for (nstp=1;nstp<=MAXSTP;nstp++) {
	if ((t+h-t2)*(t+h-t1) > 0.0) h=t2-t;
	rkqs(t,h,eps,hdid,hnext,rndm);		// deterministic part
	if (hdid == h) ++nok; else ++nbad;
	y.normalize();				// renormalize
	if ((t-t2)*(t2-t1) >= 0.0) {
		ystart = y;
		h1 = hnext;
		if( listPtr < listDim ) {	// if keeping track...
		  dtNumList[listPtr] = nok+nbad;  // ...store number of steps
		  listPtr++;
		}
		return;
	}
	if (fabs(hnext) <= hmin) {
	  cerr << "Step size too small in odeint" << endl;
	  exit(1);
	}
	h=hnext;
  }
  cerr << "Too many steps in routine odeint" << endl;
  exit(1);
}

void GenericDiff::derivs(double t, State& psi, State& dpsi,
                             ComplexRandom* rndm)
//
// Computes the change of the state psi in a timestep dt due to the
// deterministic part of a generic quantum diffusion equation
//
// Also, when stochasticFlag is set, computes the stochastic part.
//
{
  dpsi = psi;
  if (alpha != 0) {
    dpsi *= alpha(t,psi,H,nL,L);		// Deterministic term
    if( stochasticFlag != 0 ) {
      newsum = psi;
      newsum *= beta(t,psi,H,nL,L,rndm);	// Stochastic term
    }
  } else dpsi = 0;
  stochasticFlag = 0;
}

#undef MAXSTP
#undef TINY
#undef SAFETY
#undef PGROW
#undef PSHRNK
#undef ERRCON
