//   Matrix.cc -- Matrix algebra in Hilbert space.
//     
//   Copyright (C) 2007  Todd Brun, Bilal Shaw
//   
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//   
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//   
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//   ----------------------------------------------------------------------
//   If you improve the code or make additions to it, or if you have
//   comments or suggestions, please contact us:
//
//  
//
//   Dr. Todd Brun
//   Communication Sciences Institute           Tel    +001-213-740-3503
//   Department of Electrical Engineering       FAX    +001-213-740-8729 
//   University of Southern California          email  tbrun@usc.edu
//   3740 McClintock Ave, EEB 502
//   Los Angeles, CA, 90089-2565
//   USA			        
//
//   Bilal Shaw                                
//   Communication Sciences Institute          Tel    +001-213-740-5564 
//   Department of Electrical Engineering        
//   University of Southern California         email  bilalsha@usc.edu
//   3740 McClintock Ave, EEB 502
//   Los Angeles, CA, 90089-2565
//   USA			        
//
/////////////////////////////////////////////////////////////////////////////




#include <math.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <string>

#include "Matrix.h"
#include "PrimOp.h"
#include "Operator.h"
#include "CPmap.h"
using namespace std;


static double smallSize = 1.0e-8;
static int intSqrtMax = 0;
static double* intSqrtList = 0;


// This utility function is used to maintain the list of square roots.
// This function has file scope.

static void setIntSqrtMax(int maximum)
{
  int newsize,i;
  double* newPtr;

  if( maximum > intSqrtMax) {			// make sure of size
    newsize = 2*maximum;			// allow plenty of room
    newPtr = new double[newsize];		// allocate space
    for (i=0; i<intSqrtMax; i++)
      newPtr[i] = intSqrtList[i];		// re-use calculated values
    for (i=intSqrtMax; i<newsize; i++)
      newPtr[i] = sqrt(i);			// calculate new values
#ifndef NON_GNU_DELETE
    if(intSqrtList != 0) delete [] intSqrtList;	// recycle memory
#else // ---- NON GNU CODE ----
    if(intSqrtList != 0) delete [intSqrtMax] intSqrtList; // recycle memory
#endif // ---- NON GNU CODE ----
    intSqrtList = newPtr;			// point to new list	
    intSqrtMax = newsize;			// remember new size
  }
}

//Constructor for uninitialized Matrix
Matrix::Matrix():State()
{
  nFreedoms = 0;
  totalDim = 0;
  maxSize = 0;
  nDims = 0;
  sizes = 0;
  nSkips = 0;
  partDims = 0;
  freedomTypes = 0;
  coords = 0;
  data = 0;
  myPointer = 0;
  nSkip = 0;
  myType = ALL;
  mySize = 0;
  coord = 0;
}

//Constructor for 1D ground state
Matrix::Matrix(int n, FreedomType form):State()
{
  #ifndef OPTIMIZE_QSD
  
  if (n<2) error("Invalid dimension size in Matrix constructor!");
  if( (form==SPIN) && (n != 2) ) error("Spins must have dimension 2!");
  #endif
  	
  nFreedoms = 1;
  totalDim = maxSize = mySize = n;
  nSkip = 1;
  coord = 0;
  myType = form;
  data = new Complex[totalDim*totalDim];
  coords = 0;
  myPointer = data;
  nDims = 0;
  sizes = 0;
  nSkips = 0;
  partDims = 0;
  freedomTypes = 0;
  //coords = 0;
 
  
  for (int i = 0; i < totalDim; i++)
  {
    for (int j = 0; j < totalDim; j++)
    {
      data[i + totalDim*j] = 0;
      
    }
  }
	
  data[0] = 1.0;// ground state
}


// Creates a constructor of type |psi><psi|
// pure density operator

Matrix::Matrix(const State& psi):State()
{
  #ifndef OPTIMIZE_QSD
  if (psi.totalDim<2) error("Invalid dimension size in Matrix::Matrix(const State& psi) !");
  //if( (psi.myType==SPIN) && (psi.totalDim != 2) ) error("Spins must have dimension 2!");
  #endif
	
  int i, j;
  nFreedoms = psi.nFreedoms;
  
  totalDim = psi.totalDim;
  maxSize = psi.maxSize;
  mySize = psi.mySize;
  nSkip = 1;
  coord = 0;
  myType = psi.myType;
  data = new Complex[totalDim*totalDim];
  coords = 0;
  myPointer = data;
  nDims = 0;
  sizes = 0;
  nSkips = 0;
  partDims = 0;
  freedomTypes = 0;
 
  if (nFreedoms > 1)
  {
    nDims = new int[psi.nFreedoms];
    sizes = new int[psi.nFreedoms];
    nSkips = new int[psi.nFreedoms];
    partDims = new int[psi.nFreedoms];
    freedomTypes = new FreedomType[psi.nFreedoms];
    coords = new Complex[psi.nFreedoms];
    
    for(i = 0; i < psi.nFreedoms; i++)
    {
      nDims[i] = psi.nDims[i];
      sizes[i] = psi.sizes[i];
      nSkips[i] = psi.nSkips[i];
      partDims[i] = psi.partDims[i];
      freedomTypes[i] = psi.freedomTypes[i];
      coords[i] = psi.coords[i];
    }
  } 
        
 
  for(i = 0; i < mySize; i++)
  {  
    for(j = 0; j < mySize; j++)
    {  
      data[i + totalDim*j] = psi.data[i]*conj(psi.data[j]);
    }
  }
}
 
// Copy initializer
Matrix::Matrix(const Matrix& rho):State()
{
  nFreedoms = 0;
  totalDim = 0;
  copy(rho);
}

//Constructor for coherent state
Matrix::Matrix(int n, Complex alpha, FreedomType form):State()     
{							  																					
#ifndef OPTIMIZE_QSD
  if (n<2) error("Invalid dimension size in Matrix constructor!");
  if( form != FIELD )
    error("Constructor only valid for FIELD type.");
#endif
  data = 0;
  coherent(n,alpha);		// call coherent state function
  
}

// Constructor for excited coherent state
Matrix::Matrix(int n, int nstate, Complex alpha, FreedomType form)
{
#ifndef OPTIMIZE_QSD
  if (n<2) error("Invalid dimension size in Matrix constructor!");
  if( form != FIELD )
    error("Constructor only valid for FIELD type.");
#endif
  data = 0;
  fock(n,nstate);	// call Fock state function
  coord = alpha;	// displace in phase space
}

void Matrix::coherent(int n, Complex alpha)	// Create a coherent state
{
#ifndef OPTIMIZE_QSD
  if( n < 2 )
    error("Invalid dimension size in Matrix::coherent!");	 // exit on error
  if( norm(alpha) > n )
    error("Coherent state too big for storage in Matrix::coherent!");
#endif
  Complex z1;
  Complex z2;
  Complex conjalpha = conj(alpha);
  free();				// recycle memory
  totalDim = maxSize = mySize = n;	// set up state
  nFreedoms = 1;
  nSkip = 1;
  coord = 0;
  myType = FIELD;
  data = new Complex[totalDim*totalDim];		// allocate memory
  myPointer = data;
  nDims = 0;
  sizes = 0;
  nSkips = 0;
  partDims = 0;
  freedomTypes = 0;
  coords = 0;
  z1 = exp(-norm(alpha));		// normalization factor
  for (int i = 0; i < totalDim; i++) 
  {
    z2 = 1;
    for (int j = 0; j < totalDim; j++)
    {
      data[i + totalDim*j] = z1*z2;			// coherent state elements
      z2 *= conjalpha/sqrt(j+1);
    }
    z1 *= alpha/sqrt(i+1);
  }
}// end coherent state function

// Constructor for general 1 degree of freedom state
Matrix::Matrix(int n, Complex* elements, FreedomType form):State() 
{
#ifndef OPTIMIZE_QSD
  if (n<2) error("Invalid dimension size in Matrix constuctor!");
  if( (form==SPIN) && (n != 2) ) error("Spins must have dimension 2!");
#endif
  nFreedoms = 1;
  totalDim = maxSize = mySize = n;
  nSkip = 1;
  coord = 0;
  myType = form;
  data = new Complex[totalDim*totalDim];
  myPointer = data;
  nDims = 0;
  sizes = 0;
  nSkips = 0;
  partDims = 0;
  freedomTypes = 0;
  coords = 0;
  for (int i=0; i < totalDim; i++) 
	{
		for (int j = 0; j < totalDim; j++)
		{
			data[i + totalDim*j] = elements[i + totalDim*j]; 
		}
	}
}

//Constructor for multidimensional 
Matrix::Matrix(int n, int* dimensions, FreedomType* forms):State() 
{							 
  int i,j;						 // ground state
#ifndef OPTIMIZE_QSD
  if (n<1) error("Illegal number of freedoms in Matrix constructor!");
  if (n=1) error("Do not call this constructor for 1-freedom states!");
#endif
  nFreedoms = n;
  nDims = new int[n];			// allocate memory
  sizes = new int[n];
  nSkips = new int[n];
  partDims = new int[n];
  freedomTypes = new FreedomType[n];
  coords = new Complex[n];
  totalDim = 1;
  for (i=0; i<n; i++) {
#ifndef OPTIMIZE_QSD
    if(dimensions[i] < 1) error("Illegal dimension in Matrix::Matrix multidimensional constructor");
    if( (forms[i] == SPIN) && (dimensions[i] != 2) )
      error("Spin states must have 2 dimensions!");
#endif
    nSkips[i] = totalDim;
    totalDim *= (nDims[i] = dimensions[i]);
    sizes[i] = nDims[i];
    partDims[i] = totalDim;
    freedomTypes[i] = forms[i];
    coords[i] = 0;
  }
  maxSize = totalDim;
  data = new Complex[totalDim*totalDim];		// allocate memory
  myPointer = data;
  mySize = totalDim;
  nSkip = 1;
  coord = 0;
  myType = freedomTypes[0];
  
  for (i=0; i<totalDim; i++) 
    for (j=0; j < totalDim; j++)
        data[i + totalDim*j] = 0;

  data[0] = 1;
}

/*********************** BEGIN PRODUCT DENSITY MATRIX *********************************/
Matrix::Matrix(int n, Matrix* matrixList):State()
{
#ifndef OPTIMIZE_QSD
    if (n<=1) error("Illegal number of freedoms in Matrix::productMatrix constructor!");

#endif
    data = 0;
    productMatrix(n, matrixList); 	// call product density matrix function
}  

void Matrix::productMatrix(int n, Matrix* matrixList)	// Create a product state
{
  int i,j,k1,k2,l1,l2,t;
  Complex element;
#ifndef OPTIMIZE_QSD
  if( n < 1 ) error("Invalid number of degrees of freedom in productMatrix!");
#endif
  free();
  nFreedoms = n;
  nDims = new int[n];			// allocate memory
  sizes = new int[n];
  nSkips = new int[n];
  partDims = new int[n];
  freedomTypes = new FreedomType[n];
  coords = new Complex[n];
  totalDim = 1;
  for (i=0; i<n; i++) 
  {			// set up memory structure for data
    #ifndef OPTIMIZE_QSD
      if( matrixList[i].nFreedoms != 1 )
        error("Can only produce products of 1D density matrices!");
      if( matrixList[i].totalDim < 2 )
        error("Invalid dimension size in productMatrix!");
    #endif
        
    nSkips[i] = totalDim;
    totalDim *= (nDims[i] = matrixList[i].totalDim);
    sizes[i] = nDims[i];
    partDims[i] = totalDim;
    freedomTypes[i] = matrixList[i].myType;
    coords[i] = matrixList[i].coord;
  }// enf for loop
    
  mySize = totalDim;
  maxSize = totalDim;
  nSkip = 1;				// initialize state variables
  coord = coords[0];
  myType = freedomTypes[0];
  data = new Complex[totalDim*totalDim];
  myPointer = data;
    
  for (i = 0; i < totalDim; i++) 
  {  
    for (j = 0; j < totalDim; j++)
    {
      k1 = i;
      k2 = j;
      element = 1;
      for (t = n-1; t >= 0; t--) 
      {
        l1 = k1/nSkips[t];			// index i_j of jth state
        l2 = k2/nSkips[t];            
        element *= matrixList[t].data[l1+nDims[t]*l2];	// product state
        if( element == 0 ) break;		// if zero, leave this inner loop
        k1 %= nSkips[t];
        k2 %= nSkips[t];
      } // end t for loop
       data[i+totalDim*j] = element;	// psi_1[i_1]*psi_2[i_2]*psi_3[i_3]*...*psi_n[i_n]
       
    }// end j for loop
  }// end i for loop
}// end function
/************************** END PRODUCT DENSITY MATRIX **************************/


/************************** BEGIN FOCK MATRIX ***********************************/

//Constructor for Fock State

Matrix::Matrix(int n, int nstate, FreedomType form):State()	 
{							
#ifndef OPTIMIZE_QSD
  if (n<2) error("Invalid dimension size in State constructor!");
  if( form != FIELD )
    error("Constructor only valid for FIELD type.");
#endif
  data = 0;
  fock(n,nstate);		// call Fock state function
}


void Matrix::fock(int n, int nstate)	// Create a Fock State
{
#ifndef OPTIMIZE_QSD
  if( (nstate >= n) || (n < 2) || (nstate < 0) )	// check dimension
    error("Illegal arguments to Matrix::fock!");		// exit on error
#endif
  free();				// recycle memory
  totalDim = maxSize = mySize = n;	// set up state
  nFreedoms = 1;
  nSkip = 1;
  coord = 0;
  myType = FIELD;
  data = new Complex[totalDim*totalDim];		// allocate memory
  myPointer = data;
  nDims = 0;
  sizes = 0;
  nSkips = 0;
  partDims = 0;
  freedomTypes = 0;
  coords = 0;
  for (int i = 0; i < totalDim; i++) 
  {
    for (int j = 0; j < totalDim; j++)
    {
      data[i + totalDim*j] = 0;
    }
  }
	
  data[nstate + totalDim*nstate] = 1;					// Fock state
}
/*************************** END FOCK MATRIX *************************************/


/*************************** BEGIN GAUSSIAN MATRIX ******************************/

//Constructor for Gaussian State

Matrix::Matrix(int n, Complex alpha, double width, double dx, FreedomType form):State()
{					// Constructor for Gaussian state
#ifndef OPTIMIZE_QSD
  if (n<2) error("Invalid dimension size in Matrix constructor!");
  if( form != PARTICLE )
    error("Constructor only valid for PARTICLE type.");
#endif
  data = 0;
  gaussian(n,alpha,width,dx);		// call Gaussian state function
}


void Matrix::gaussian(int n, Complex alpha, double width, double dx)
{					// Create a Gaussian state
// The PARTICLE type is a position representation.  It stores the center
// of coordinates in the real part of coord.  In the imaginary part, it
// stores the lattice spacing dx.  If imag(coord) == 0, then it assumes
// that dx=1.  The amplitudes are stored in the data array with the
// negative locations in the odd elements and the nonnegative in the
// even elements.
#ifndef OPTIMIZE_QSD
  if( n < 2 )
    error("Invalid state size in Matrix::gaussian!");	 // exit on error
  if( dx < 0 )
    error("Invalid lattice spacing in Matrix::gaussian!"); // exit on error
  if( (dx != 0) && (width/dx > n) )
    error("Gaussian state too big for storage in Matrix::gaussian!");
  if( (dx == 0) && (width > n) )
    error("Gaussian state too big for storage in Matrix::gaussian!");
#endif
  Complex p(0.0,0.5*alpha.im*dx);	// center of Gaussian in momentum
  double coeff = 0.25*dx*dx/(width*width);	// coefficient in exponent
  Complex phase(0.0,alpha.im*alpha.re);
  free();				// recycle memory
  totalDim = maxSize = mySize = n;	// set up state
  nFreedoms = 1;
  nSkip = 1;
  Complex z(alpha.re,dx);
  coord = z;				// center of Gaussian, lattice space
  myType = PARTICLE;
  data = new Complex[totalDim*totalDim];		// allocate memory
  myPointer = data;
  nDims = 0;
  sizes = 0;
  nSkips = 0;
  partDims = 0;
  freedomTypes = 0;
  coords = 0;
  int i, j;
  
  // even-even
  for(i = 0; i < totalDim; i += 2)
  {
    for(j = 0; j < totalDim; j += 2)
    {
      data[i + totalDim*j] = exp(-coeff*(i*i + j*j) + p*(i - j));
    }
  }
  
  // even-odd
  
  for(i = 0; i < totalDim; i +=2)
  {
    for(j = 2; j < totalDim+1; j += 2)
    {
      data[i + totalDim*(j-1)] = exp(-coeff*(i*i + j*j) + p*(i + j));
    }
  }
  
  // odd-even
  
  for(i = 2; i < totalDim+1; i += 2)
  {
    for(j = 0; j < totalDim; j += 2)
    {
      data[(i-1) + totalDim*j] = exp(-coeff*(i*i + j*j) - p*(i + j));
    }
  }
  
  // odd-odd
  
  for(i = 2; i < totalDim+1; i += 2)
  {
    for(j = 2; j < totalDim+1; j += 2)
    {
      data[(i-1) + totalDim*(j-1)] = exp(-coeff*(i*i + j*j) - p*(i - j));
    }
  }
  normalize();
}
/*************************** END GAUSSIAN MATRIX ************************************/

Matrix::~Matrix()
{
  free();
  
}
	
void Matrix::free()
{
  if(data != 0)
  {
#ifndef NON_GNU_DELETE
    delete[] data;
    if(nFreedoms > 1)
    {
      delete[] nDims;
      delete[] sizes;
      delete[] nSkips;
      delete[] partDims;
      delete[] freedomTypes;
      delete[] coords;
    }
#else // ---- NON GNU CODE ----
    delete [totalDim*totalDim] data;		// NON GNU recycle storage
    if( nFreedoms > 1 ) {
      delete [nFreedoms] nDims;		// NON GNU recycle parameter arrays
      delete [nFreedoms] sizes;
      delete [nFreedoms] nSkips;
      delete [nFreedoms] partDims;
      delete [nFreedoms] freedomTypes;
      delete [nFreedoms] coords;
    }
#endif // ---- NON GNU CODE ----
    nFreedoms = 0;			// zero variables
    totalDim = 0;
    maxSize = 0;
    myType = ALL;
    mySize = 0;
    nSkip = 0;
    coord = 0;
    myPointer = 0;			// reset pointers to zero
    data = 0;
    nDims = 0;
    sizes = 0;
    nSkips = 0;
    coords = 0;
    partDims = 0;
    freedomTypes = 0;
  }
}
	

void Matrix::normalize()
{
  
  int i,j;
  Complex trace = this->Tr();
  
  
  if (trace == 0)
    error("Division by zero trace not allowed Matrix::normalize()"); 
  
  for (i = 0; i < maxSize; i++)
  {
    for (j = 0; j < maxSize; j++)
    {
      data[i + maxSize*j] /= trace ;
    }
  }
  
}

void Matrix::xerox(Matrix& rho)
// Unlike copy, xerox makes a copy of an existing matrix including only
// the memory which is actually being used, rather than all the memory
// allocated.  This is useful as a more efficient way of creating
// temporary matrices than using the normal copy-constructor.
//
// Note that xerox should NOT be used to assign to a matrix which already
// has memory allocated.  This results in inefficient memory handling, and
// might cause errors.
{
  int i,j;
#ifndef OPTIMIZE_QSD
  if( data != 0 )
    error("Matrix::xerox should not be used on a matrix with allocated memory!");
#endif
  nFreedoms = rho.nFreedoms;
  if( nFreedoms > 1 ) {
    nDims = new int[nFreedoms];
    sizes = new int[nFreedoms];
    nSkips = new int[nFreedoms];
    partDims = new int[nFreedoms];
    freedomTypes = new FreedomType[nFreedoms];
    coords = new Complex[nFreedoms];
  }
  totalDim = rho.maxSize;
  data = new Complex[totalDim*totalDim];
  maxSize = rho.maxSize;			// copy member data
  nSkip = rho.nSkip;
  coord = rho.coord;
  mySize = rho.mySize;
  myType = rho.myType;
  myPointer = data;
  for (i=0; i < maxSize; i++)
    for (j = 0; j < maxSize; j++)
      data[i + maxSize*j] = rho.data[i + maxSize*j];
  
  if( nFreedoms > 1 ) {
    for (i=0; i<nFreedoms; i++) {
      nDims[i] = rho.sizes[i];
      sizes[i] = rho.sizes[i];
      nSkips[i] = rho.nSkips[i];
      partDims[i] = rho.partDims[i];
      freedomTypes[i] = rho.freedomTypes[i];
      coords[i] = rho.coords[i];
    }
  }
}

Matrix& Matrix::operator=(const Matrix& rho) //assignment operator
{
  
  if (this != &rho)// otherwise rho = rho would lead to trouble
  {
    if((totalDim != rho.totalDim) || (nFreedoms != rho.nFreedoms))
    {
      free();//recycle any memory
    }
		
   copy(rho);//copy rho into b
  }
	
  return *this;
}

void Matrix::copy(const Matrix& rho)
{
  int i,j;
  if( nFreedoms != rho.nFreedoms)
  {
    nFreedoms = rho.nFreedoms;
    if( nFreedoms > 1)
    {
      nDims = new int[nFreedoms];
      sizes = new int[nFreedoms];
      nSkips = new int[nFreedoms];
      partDims = new int[nFreedoms];
      freedomTypes = new FreedomType[nFreedoms];
      coords = new Complex[nFreedoms];
    }
  }
	
  if( totalDim != rho.totalDim ) 
  {	// if no memory allocated, allocate it
    totalDim = rho.totalDim;
    data = new Complex[totalDim*totalDim];
  }
  maxSize = rho.maxSize;			// copy member data
  nSkip = rho.nSkip;
  coord = rho.coord;
  mySize = rho.mySize;
  myType = rho.myType;
  myPointer = data;
  for (i = 0; i < maxSize; i++)
  {
    for (j = 0; j < maxSize; j++)
    {
      data[i + maxSize*j] = rho.data[i + maxSize*j];
    
    }
  }
  if( nFreedoms > 1 ) 
  {
   for (i = 0; i < nFreedoms; i++) 
   {
      nDims[i] = rho.nDims[i];
      sizes[i] = rho.sizes[i];
      nSkips[i] = rho.nSkips[i];
      partDims[i] = rho.partDims[i];
      freedomTypes[i] = rho.freedomTypes[i];
      coords[i] = rho.coords[i];
    }
  }
}
	
Matrix& Matrix::operator=(int n)
{
  if(n != 0)
    error("Cannot assign a Matrix a value other than zero!");
  for(int i = 0; i < maxSize; i++)
  {
    for(int j = 0; j < maxSize; j++)
    {
      data[i + maxSize*j] = 0;
    }
  }
	
  return *this;
}			 

/*
Complex* Matrix::operator[](int i)
{
  
  return myPointer+totalDim*i;
}
*/


// This will send a matrix rho to A_j*rho*A_j_dagger + ....
Matrix& Matrix::operator*=(CPmap& map)
{
  int size = map.getcpmapSize();
  Matrix* phi = new Matrix(this->mySize,this->myType);
  Matrix* sum = new Matrix(this->mySize,this->myType);
   
  for(int i = 0; i < phi->mySize; i++){
    for(int j = 0; j < phi->mySize; j++){
      phi->data[i + mySize*j] = 0;
      sum->data[i + mySize*j] = 0;
    }
  }
   
  Operator* cpmap = map.getcpmapPointer();
    
  for(int j = 0; j < size; j++)
  {
    
    *phi = cpmap[j]*(*this);//A_j*this
    phi->hc(); // this_dagger*A_j_dagger
    *phi *= cpmap[j]; //A_j*this_dagger*A_j_dagger
    phi->hc();// in case phi is not Hermitian
    *sum += *phi;
    
  }
 
  *this = *sum;
  
  return *this;
}
  
Matrix& Matrix::operator*=(const Matrix& rho)
{
  int i,j,k;
  if(mySize != rho.mySize)
    error("Incompatible matrix sizes in Matrix::operator*!");
  if(myType != rho.myType)
    error("Incompatible matrix types in Matrix::operator*!");
	
  Complex* sum = new Complex[mySize*mySize];
 
  for(i = 0; i < mySize; i++)
  {
    for(j = 0; j < mySize; j++)
    {
      sum[i + mySize*j] = 0;
      for(k = 0; k < mySize; k++)
        sum[i + mySize*j] += rho.data[i+k*mySize]*data[k+j*mySize];
    }// end i loop
    
  }// end j loop	

  for( i = 0; i < mySize*mySize; i++)
  {
    data[i] = sum[i];
  }

#ifndef NON_GNU_DELETE
  delete[] sum;
#else
  delete[mySize*mySize] sum;
#endif
  
  return *this;
}

Matrix& Matrix::operator*=(const Complex& z)
{
  for (int i = 0; i < mySize; i++)
  {
    for (int j = 0; j < mySize; j++)
    {
      data[i + mySize*j] = (data[i + mySize*j])*z;
    }
  }
  
  return *this;
}

Matrix& Matrix::operator*=(double x)
{
  for (int i = 0; i < mySize; i++)
  {
    for (int j = 0; j < mySize; j++)
    {
      data[i + mySize*j] *= x;
    }
  }
  
  return *this;
}

Matrix& Matrix::operator*=(ImaginaryUnit im)
{
  int i,j;
  if(im == IM)
  {
    for(i = 0; i < mySize; i++)
    {
      for(j = 0; j < mySize; j++)
      {
        data[i + mySize*j].timesI();
      }
    }
  }
  if (im == M_IM)
  {
    for(i = 0; i < mySize; i++)
    {
      for(j = 0; j < mySize; j++)
      {
        data[i + mySize*j].timesMinusI();
      }
    }
  }

  return *this;
}


Matrix& Matrix::operator+=(const Matrix& rho)
{
  if(mySize != rho.mySize)
    error("Incompatible matrix sizes in Matrix::operator+!");
  if(myType != rho.myType)
    error("Incompatible matrix types in Matrix::operator+!");
  for(int i = 0; i < mySize; i++)
  {
    for(int j = 0; j < mySize; j++)
    {
      data[i + mySize*j] = data[i + mySize*j] + rho.data[i + mySize*j];
    }	  
  }
	
  return *this;
}


Matrix& Matrix::operator-=(const Matrix& rho)
{
  
  if(mySize != rho.mySize)
    error("Incompatible matrix sizes in Matrix::operator-!");
  if(myType != rho.myType)
    error("Incompatible matrix types in Matrix::operator-!");
  for(int i = 0; i < mySize; i++)
  {
    for(int j = 0; j < mySize; j++)
    {
      data[i + mySize*j] = data[i + mySize*j] - rho.data[i + mySize*j];
    }	  
  }
	
  return *this;
}

//Partial trace of a matrix.  Traces out over the specified degree of freedom
// and returns the reduced density matrix over all the other non-traced out
// degrees of freedom.

Matrix Matrix::reducedMatrix(int theFreedom)
{
  
  Matrix mytemp(nDims[theFreedom],freedomTypes[theFreedom]);
  int i,j,k,m;
  //one degree of freedom case
  if (nFreedoms == 1)
    mytemp = *this;
  
  else //multi-freedom case
  {
        
    mytemp.coord = coords[theFreedom]; 
    mytemp.mySize = sizes[theFreedom];
    mytemp.maxSize = sizes[theFreedom];
    mytemp.nSkip = 1;
    
    for(i = 0; i < sizes[theFreedom]; i++)
      for(j = 0; j < sizes[theFreedom]; j++)
      {
        mytemp.data[i + j*sizes[theFreedom]] = 0;
        for(k = 0; k < nSkips[theFreedom]; k++)//inner
          for(m = 0; m < maxSize; m += partDims[theFreedom])//outer
            mytemp.data[i + j*sizes[theFreedom]] += data[nSkips[theFreedom]*(i +
                j*maxSize) + (k+m)*(maxSize + 1)];
      }

  }
  return mytemp;
  
}


//Trace of a matrix
Complex Matrix::Tr()
{
  
  int i;
  Complex sum = 0.0;
	
  for(i = 0; i < maxSize; i++)
  {
    sum += data[i + maxSize*i];
  }
	
  return sum;
}
	

//Transpose of a matrix
Matrix& Matrix::transpose()
{
  int i,j;
  Complex temp;
  
  for (i = 0;i < mySize;i++)
    for(j = 0; j < i; j++)
    {   
       temp = data[i + mySize*j];
       data[i + mySize*j] = data[j + mySize*i];
       data[j + mySize*i] = temp;
     }  
 
  return *this;
}

//Conjugate of a matrix
Matrix& Matrix::conjugate()
{
  int i,j;
  
      
  for(i = 0; i < mySize; i++)
  {
    for(j = 0; j < mySize; j++)
    {
      
      data[i + mySize*j] = conj(data[i + mySize*j]);
    }
  }
  return *this;
}


Matrix& Matrix::hc()
{
  int i,j;
  Complex temp;
   
  for(i = 0; i < mySize; i++)
  {
    data[i + mySize*i] = conj(data[i + mySize*i]);
  }  
  
  for (i = 0;i < mySize;i++)
    for(j = 0; j < i; j++)
    {   
       temp = conj(data[i + mySize*j]);
       data[i + mySize*j] = conj(data[j + mySize*i]);
       data[j + mySize*i] = temp;
     }  
       
  return *this;
}	

	
//Output a matrix to an output stream
ostream& operator<<( ostream& s, const Matrix& rho)
{
  int i, j;
  s << rho.nFreedoms << endl;
  s << rho.totalDim << endl;
  s << rho.maxSize << endl;
  s << rho.mySize << endl;
  s << rho.nSkip << endl;
  s << rho.coord << endl;
  s << rho.myType << endl;
  if(rho.nFreedoms > 1)
  {
    for (i = 0; i < rho.nFreedoms; i++)
    {
      s << rho.nDims[i] << endl;
      s << rho.sizes[i] << endl;
      s << rho.nSkips[i] << endl;
      s << rho.partDims[i] << endl;
      s << rho.freedomTypes[i] << endl;
      s << rho.coords[i] << endl;
    }
    for (i = 0; i < rho.maxSize; i++)
    {
      for (j = 0; j < rho.maxSize; j++)
      {
        s << rho.data[i*rho.maxSize + j] << endl;
      }
    }
  }
	
  return s;
} 

istream& operator>>( istream& s, Matrix& rho )
//
// Input a State from an input stream
//
{
  int i,j;

  rho.free();			// free up State
  s >> rho.nFreedoms;
  s >> rho.totalDim;
  s >> rho.maxSize;
  s >> rho.mySize;
  s >> rho.nSkip;
  s >> rho.coord;
  s >> rho.myType;
  if( rho.totalDim > 0 ) {
    rho.data = new Complex[rho.totalDim*rho.totalDim];
    rho.myPointer = rho.data;
  }
  if( rho.nFreedoms > 1 ) {
    rho.nDims = new int[rho.nFreedoms];
    rho.sizes = new int[rho.nFreedoms];
    rho.nSkips = new int[rho.nFreedoms];
    rho.partDims = new int[rho.nFreedoms];
    rho.freedomTypes = new FreedomType[rho.nFreedoms];
    rho.coords = new Complex[rho.nFreedoms];
    for(i=0; i<rho.nFreedoms; i++) 
		{
      	s >> rho.nDims[i];
      	s >> rho.sizes[i];
      	s >> rho.nSkips[i];
      	s >> rho.partDims[i];
      	s >> rho.freedomTypes[i];
      	s >> rho.coords[i];
    }
  }
  
  for (i = 0; i < rho.maxSize; i++)
  {
    for(j = 0; j < rho.maxSize; j++)
    { 	
      s >> rho.data[i + rho.maxSize*j];    }
  }
  
  return s;
}

void Matrix::apply(PrimaryOperator& theOp, int hc, int theFreedom, FreedomType theType, double t)
{
  
  int i,j,numcol;
  if( nFreedoms == 1 )// 1 freedom case 
  {		
    #ifndef OPTIMIZE_QSD
      if( (theType != myType) && (theType != ALL) )
        error("Incompatible operator type in Matrix::apply!");
    #endif
    for (numcol = 0; numcol < mySize; numcol++)
    {
       myPointer = data + numcol*mySize;
       theOp.applyTo(*this,hc,t);
    }
  } // end if
  else // multi-freedom case
  {				
    
    #ifndef OPTIMIZE_QSD
      if( (theFreedom < 0) || (theFreedom >= nFreedoms) )
        error("Illegal degree of freedom requested in State::apply!");
      if( (theType != freedomTypes[theFreedom]) && (theType != ALL) )
        error("Incompatible operator type in State::apply!");
    #endif
    
    mySize = sizes[theFreedom];		// set up 1D state for PrimaryOperator
    nSkip = nSkips[theFreedom];   
    coord = coords[theFreedom];
    for (numcol = 0; numcol < maxSize; numcol++)// loop over columns of matrix
    {
      for (i = 0; i < nSkip; i++)// loop over inner and 
      {				
        for (j = 0; j < maxSize; j += partDims[theFreedom])// outer indices 
        {	
          myPointer = data + i + j + numcol*maxSize;
          if( i+j > maxSize ) 
            error("Gone off end of array!");
          theOp.applyTo(*this,hc,t);
        } 
        
      }
    }
    
    mySize = maxSize;
    nSkip = 1;
    coord = coords[0];
    myPointer = data;
  }// end else
}// end apply()

//rho*A[k]
Matrix operator*(Matrix& rho, CPmap& A)
{
  
  Matrix result(rho);
  return(result *= A);
}

// A[k]*rho
Matrix operator*(CPmap& A, Matrix& rho)
{
  
  Matrix result(rho);
  return(result *= A);
}
     
Matrix operator*(const Matrix& rho, const Matrix& phi)
{
	Matrix result(rho);
	return (result *= phi);
}

		
Matrix operator*(const Complex& z, const Matrix& rho)
{
  Matrix result(rho);
  return (result *= z);
}

Matrix operator*(const Matrix& rho, const Complex& z)
{
  Matrix result(rho);
  return (result *= z);
}

Matrix operator*(double x, Matrix& rho)
{
  Matrix result(rho);
  return (result *= x);
}

Matrix operator*(const Matrix& rho, double x)
{
  Matrix result(rho);
  return (result *= x);
}

Matrix operator*(ImaginaryUnit i, const Matrix& rho)
{
  Matrix result(rho);
  return (result *= i);
}

Matrix operator*(const Matrix& rho, ImaginaryUnit i)
{
  Matrix result(rho);
  return (result *= i);
}

Matrix operator+(const Matrix& rho, const Matrix& phi)
{
  Matrix result(rho);
  return (result += phi);
}
	
Matrix operator-(const Matrix& rho, const Matrix& phi)
{
  Matrix result(rho);
  return (result -= phi);
} 

Matrix operator+(const Matrix& rho)
{
  Matrix result(rho);
  return result;
}

Matrix operator-(const Matrix& rho)
{
  Matrix result(rho);
   
  int getmaxSize = result.getmaxSize();
  for (int i = 0; i < getmaxSize; i++)
  {
    for (int j = 0; j < getmaxSize; j++)
    {
      result.data[i*getmaxSize + j] = -result.data[i*getmaxSize + j];
    }
  }
	
  return result;
}

int Matrix::size()
{
  return mySize;
}

int Matrix::getSize(int theFreedom)
{
  #ifndef OPTIMIZE_QSD
  if( (theFreedom < 0) || (theFreedom >= nFreedoms) )
    error("Illegal freedom passed to Matrix::getSize!");
#endif
  return sizes[theFreedom];
}


void Matrix::diagnostic()	// for debugging
{
  cout << "I AM IN MATRIX" << endl;
  int i;
  cout << "Number of Freedoms " << nFreedoms << ".\n";
  cout << "Total Dimensions " << totalDim << ".\n";
  cout << "MaxSize " << maxSize << ".\n";
  cout << "My Type " << myType << ".\n";
  cout << "My Size " << mySize << ".\n";
  cout << "nSkip " << nSkip << ".\n";
  cout << "coord " << coord << ".\n";
  cout << "data pointer " << data << ".\n";
  cout << "My Pointer " << myPointer << ".\n";
  cout << "Dimensions:\n";
  if (nFreedoms > 1 ) {
    for (i=0; i<nFreedoms; i++)
      cout << "  " << nDims[i] << ".\n";
    cout << "Sizes:\n";
    for (i=0; i<nFreedoms; i++)
      cout << "  " << sizes[i] << ".\n";
    cout << "Skips:\n";
    for (i=0; i<nFreedoms; i++)
      cout << "  " << nSkips[i] << ".\n";
    cout << "Partial dimensions:\n";
    for (i=0; i<nFreedoms; i++)
      cout << "  " << partDims[i] << ".\n";
    cout << "Centers of coordinates:\n";
    for (i=0; i<nFreedoms; i++)
      cout << "  " << coords[i] << ".\n";
  }

}
	
//////////////// BASIS CHANGING PART ADDED HERE////////////////////////////
//////////////////////////////////////////////////////////////////////////

Complex Matrix::centerCoords()
{
  return coord;
}


Complex Matrix::getCoords(int theFreedom)
{
  #ifndef OPTIMIZE_QSD
  if( (theFreedom<0) || (theFreedom >= nFreedoms) )
    error("Illegal freedom requested in Matrix::getCoords!");
#endif
  if( nFreedoms > 1 ) return coords[theFreedom];
  return coord;
}

void Matrix::moveStep(Complex& z1, Complex& z2)
{}

void Matrix::stretchFreedom(int theFreedom, int theSize)
//
// If the space being used for a degree of freedom is not sufficient to
// avoid truncation errors, stretch the available memory.
//
{
  int i,j,k,numcol;
  int newDim,newSize;

// Stretch the data into a larger space
  
  newDim = nSkips[theFreedom]*theSize;		// new partial dimension
  newSize = theSize*maxSize/sizes[theFreedom];	// new total size
  int newSizesquare = newSize*newSize;
  
  for(i = maxSize*newSize; i < newSizesquare; i++)
    data[i] = 0;
  
  for(numcol = maxSize-1; numcol >= 0; numcol--)
  {
    for(i = maxSize-partDims[theFreedom], j = newSize-newDim; i >= 0;
        i -= partDims[theFreedom], j -=newDim){
        for(k = partDims[theFreedom]-1; k>=0;k--)
          data[j+k+numcol*newSize] = data[i+k+numcol*maxSize];
        for(k = newDim-1;k>=partDims[theFreedom];k--)
          data[j+k+numcol*newSize] = 0;
      }
  }
  
// Recompute nSkips and partDims
  sizes[theFreedom] = theSize;			// size of freedom
  partDims[theFreedom] = newDim;
  maxSize = newSize;				// total size used
  for(i=theFreedom+1; i<nFreedoms; i++) {
    nSkips[i] = partDims[i-1];			// new nSkips
    partDims[i] = nSkips[i]*sizes[i];		// new partDims
  }
  mySize = maxSize;
}

void Matrix::shrinkFreedom(int theFreedom, int theSize)
//
// If more memory is being used than is necessary to avoid truncation
// error, shrink memory used by freedom.
//
{
  int i,j,k,numcol;
  int newDim,newSize;

// Compact the data into a smaller space
  newDim = nSkips[theFreedom]*theSize;		// new partial dimension
  newSize = theSize*maxSize/sizes[theFreedom];	// new total size
  for(numcol = 0; numcol < newSize; numcol++)
  for(i=0,j=0; i<maxSize; i+=partDims[theFreedom],j+=newDim) //old and new outer indices
  for(k=0; k<newDim; k++)			// inner indices
    data[j+k+numcol*newSize] = data[k+i+numcol*maxSize];			// compact data
// Pad out the empty space with zeros
 /* 
  for(i=newSize; i<maxSize; i++) 
    for(j = newSize; j < maxSize; j++)
      data[i + maxSize*j] = 0;
*/

// Recompute nSkips and partDims
  sizes[theFreedom] = theSize;			// size of freedom
  partDims[theFreedom] = newDim;
  maxSize = newSize;				// total size used
  for(i=theFreedom+1; i<nFreedoms; i++) {
    nSkips[i] = partDims[i-1];			// new nSkips
    partDims[i] = nSkips[i]*sizes[i];		// new partDims
  }
  mySize = maxSize;
}

void Matrix::setCoords(Complex& alpha, int theFreedom)
{
  #ifndef OPTIMIZE_QSD
  if( (theFreedom<0) || (theFreedom >= nFreedoms) )
    error("Illegal freedom requested in State::setCoords!");
#endif
  if( nFreedoms == 1 )
    coord = alpha;
  else {
    coords[theFreedom] = alpha;
    if( theFreedom == 0 ) coord = alpha;
  }  
}

void Matrix::displaceCoords(Complex& alpha, int theFreedom)
//
// Reset the value of the phase space coordinate for a freedom.
{
#ifndef OPTIMIZE_QSD
  if( (theFreedom<0) || (theFreedom >= nFreedoms) )
    error("Illegal freedom requested in State::setCoords!");
#endif
  if( nFreedoms == 1 )
    coord += alpha;
  else {
    coords[theFreedom] += alpha;
    if( theFreedom == 0 ) coord += alpha;
  }  
}

double Matrix::checkBounds(int theFreedom, int numChecks)
//
// Check how much probability is in the top numChecks levels
// of the degree of freedom theFreedom.
//
{
#ifndef OPTIMIZE_QSD
  if( (theFreedom < 0) || (theFreedom >= nFreedoms) )
    error("Illegal freedom passed to Matrix::checkBounds!");
  if( (nFreedoms == 1) && (numChecks > maxSize) )
    error("Too many checks requested in Matrix::checkBounds!");
  if( (nFreedoms > 1) && (numChecks > sizes[theFreedom]) )
    error("Too many checks requested in Matrix::checkBounds!");
#endif
  Complex sum;
  int i,j,k,l;
  if( nFreedoms == 1) {				// 1 freedom case
    for(i=1; i<=numChecks; i++) {
      sum += data[(maxSize - i) + (maxSize - i)*maxSize];		// add up probability
    }
  }
  else {					// multi-freedom case
    int theSkip = nSkips[theFreedom];
    int theBound = (sizes[theFreedom] - 1)*theSkip;
    int theDim = partDims[theFreedom];
    for(i=0; i<theSkip; i++)			// loop over other indices
    for(j=0; j<maxSize; j+=theDim)
    for(k=0,l=theBound; k<numChecks; k++,l-=theSkip) {
      sum += data[(i+j+l)+(maxSize*(i+j+l))];			// add up probability
    }
  }
  return sum.re; // not sure about this
}

void Matrix::moveCoords2(const Complex& alpha, int theFreedom)
{
  
  #ifndef OPTIMIZE_QSD
  if( (theFreedom >= nFreedoms) || (theFreedom < 0) )
    error("Illegal freedom passed to State::moveCoords!");
#endif
  Complex mag = this->Tr();
  if( mag == 0 ) {				// move zero state
    if( nFreedoms == 1 ) coord -= alpha;
    if( nFreedoms > 1 ) {
      coords[theFreedom] -= alpha;
      coord = coords[0];
    }
    return;
  }
  
  int i,j,k,l,m,numcol;
  double normalpha;
  
  normalpha = norm(alpha);
  Matrix dummy1 = (*this);
  Matrix dummy2 = (*this);
  
  Complex malphacc = - conj(alpha);
  
  if (nFreedoms == 1) {				// 1 freedom case
#ifndef OPTIMIZE_QSD
  if( myType != FIELD) error("Can only move coordinates for FIELD type.");
#endif
  if( mySize > intSqrtMax) setIntSqrtMax(mySize);
    for (m=1; m<mySize-1; m++)// Calculate exp(-alpha^* a) 
    {   
      for (numcol = 0; numcol < mySize; numcol++)// loop over columns of rho
      {
         dummy2.myPointer = dummy2.data + numcol*mySize;
        
         for (k=1; k<=mySize-m; k++)	// apply local annihilation op.
         {  
            dummy2.myPointer[k-1] = intSqrtList[k]*dummy2.myPointer[k];
         }
      
         dummy2.myPointer[mySize-m] = 0;
      }
      
      
      dummy2 *= (malphacc/m);	// (-alpha^* a)^m|psi>/(m!)
      dummy1 += dummy2;
    }
    
    dummy2 = dummy1;
    
    for (m=1; m<mySize-1; m++)// Calculate exp(alpha adag) 
    {   
      for (numcol = 0; numcol < mySize; numcol++)// loop over columns of rho
      {
        dummy2.myPointer = dummy2.data + numcol*mySize;
        
        for (k=mySize-1; k>=m; k--)	// apply local creation op.
        {  
           
           dummy2.myPointer[k] = intSqrtList[k]*dummy2.myPointer[k-1];
        }     
      
        dummy2.myPointer[m-1] = 0;  
      }
      
      
      dummy2 *= (alpha/m);	// (alpha adag)^m|psi>/(m!)
      dummy1 += dummy2;
    }
    
    //dummy1 *= exp(-normalpha/2);		// exp(-|alpha|^2/2) factor
     dummy1.normalize();
    
    dummy1.hc(); // rho = rho*Ddag
    dummy2 = dummy1;
    // rho = D*rho*Ddag
    //Now we apply D to (rho*Ddag)
  if( mySize > intSqrtMax) setIntSqrtMax(mySize);
    for (m=1; m<mySize-1; m++)// Calculate exp(-alpha^* a) 
    {   
      for (numcol = 0; numcol < mySize; numcol++)// loop over columns of rho
      {
        dummy2.myPointer = dummy2.data + numcol*mySize;
        
        for (k=1; k<=mySize-m; k++)	// apply local annihilation op.
        { 
          
          dummy2.myPointer[k-1] = intSqrtList[k]*dummy2.myPointer[k];
        }              
        dummy2.myPointer[mySize-m] = 0;
        
      }
      
      
      dummy2 *= (malphacc/m);	// (-alpha^* a)^m|psi>/(m!)
      dummy1 += dummy2;
    }
    
    dummy2 = dummy1;
    
    for (m=1; m<mySize-1; m++)// Calculate exp(alpha adag) 
    {   
      for (numcol = 0; numcol < mySize; numcol++)// loop over columns of rho
      {
        dummy2.myPointer = dummy2.data + numcol*mySize;
        
        for (k=mySize-1; k>=m; k--)	// apply local creation op.
        {  
          
          dummy2.myPointer[k] = intSqrtList[k]*dummy2.myPointer[k-1];
        }
      
        dummy2.myPointer[m-1] = 0;
      }
      
      
      dummy2 *= (alpha/m);	// (alpha adag)^m|psi>/(m!)
      dummy1 += dummy2;
    }
    
    //dummy1 *= exp(-normalpha/2);
     dummy1.normalize();    
    (*this) = dummy1;
    myPointer = data;
    coord -= alpha;
  }// end single degree freedom

  else {					// multi-freedom case
#ifndef OPTIMIZE_QSD
    if( freedomTypes[theFreedom] != FIELD )
      error("Can only move coordinates for FIELD type.");
#endif
    mySize = sizes[theFreedom];
    nSkip = nSkips[theFreedom];
    if( mySize > intSqrtMax) setIntSqrtMax(mySize);
    
    for (m=1; m<mySize-1; m++)// Calculate exp(-alpha^* a) 
    {   
      for (numcol = 0; numcol < maxSize; numcol++)// loop over columns of rho
      {
        for (i=0; i<nSkip; i++)		// apply local annihilation op
        {          
          for (j=0; j<maxSize; j+=partDims[theFreedom]) 
          {	
             dummy2.myPointer = dummy2.data + i + j + numcol*maxSize;		// indices
             for (k=1,l=nSkip; k<=mySize-m; k++,l+=nSkip)
             {
                dummy2.myPointer[l-nSkip] = intSqrtList[k]*dummy2.myPointer[l];
             }
              
             dummy2.myPointer[(mySize-m)*nSkip] = 0;
          }// end j
        }// end i
      }// end loop over columns
      
      dummy2 *= (malphacc/m);	// (-alpha^* a)^m|psi>/(m!)
      dummy1 += dummy2;
    } // end m
    
    dummy2 = dummy1;
    
    for (m=1; m<mySize-1; m++)// Calculate exp(alpha adag) 
    {   
      for (numcol = 0; numcol < maxSize; numcol++)// loop over columns of rho
      {
        for (i=0; i<nSkip; i++)		// apply local creation op
        {
          for (j=0; j<maxSize; j+=partDims[theFreedom]) 
          {	
            dummy2.myPointer = dummy2.data + i + j + numcol*maxSize;		// indices
            for (k=mySize-1,l=(mySize-1)*nSkip; k>=m; k--,l-=nSkip)
            {
              dummy2.myPointer[l] = intSqrtList[k]*dummy2.myPointer[l-nSkip];
            }
            dummy2.myPointer[(m-1)*nSkip] = 0;
          }//end j
        }// end i
      }// end loop over columns
      
      dummy2 *= (alpha/m);	// (alpha adag)^m|psi>/(m!)
      dummy1 += dummy2;
    }// end m
    
    
    dummy1.hc();
    dummy1.normalize();
    dummy2 = dummy1;
    // Now we apply D to rho*Ddag    
    mySize = sizes[theFreedom];
    nSkip = nSkips[theFreedom];
    if( mySize > intSqrtMax) setIntSqrtMax(mySize);
    
    for (m=1; m<mySize-1; m++)// Calculate exp(-alpha^* a) 
    {   
      for (numcol = 0; numcol < maxSize; numcol++)// loop over columns of rho
      {
        for (i=0; i < nSkip; i++)		// apply local annihilation op
        {          
          for (j=0; j < maxSize; j+=partDims[theFreedom]) 
          {	
             dummy2.myPointer = dummy2.data + i + j + numcol*maxSize;		// indices
             for (k=1,l=nSkip; k<=mySize-m; k++,l+=nSkip)
             {
                dummy2.myPointer[l-nSkip] = intSqrtList[k]*dummy2.myPointer[l];
             }
              
             dummy2.myPointer[(mySize-m)*nSkip] = 0;
          }// end j
        }// end i
      }// end loop over columns
      
      dummy2 *= (malphacc/m);	// (-alpha^* a)^m|psi>/(m!)
      dummy1 += dummy2;
    } // end m
    
    dummy2 = dummy1;
    
    for (m=1; m<mySize-1; m++)// Calculate exp(alpha adag) 
    {   
      for (numcol = 0; numcol < maxSize; numcol++)// loop over columns of rho
      {
        for (i=0; i < nSkip; i++)		// apply local creation op
        {
          for (j=0; j < maxSize; j+=partDims[theFreedom]) 
          {	
            dummy2.myPointer = dummy2.data + i + j + numcol*maxSize;		// indices
            for (k=mySize-1,l=(mySize-1)*nSkip; k>=m; k--,l-=nSkip)
            {
              dummy2.myPointer[l] = intSqrtList[k]*dummy2.myPointer[l-nSkip];
            }
            dummy2.myPointer[(m-1)*nSkip] = 0;
          }//end j
        }// end i
      }// end loop over columns
      
      dummy2 *= (alpha/m);	// (alpha adag)^m|psi>/(m!)
      dummy1 += dummy2;
    }// end m
    
    //dummy1 *= exp(-normalpha);
    dummy1.normalize();
    (*this) = dummy1;
    myPointer = data;
    mySize = maxSize;
    nSkip = 1;
    coords[theFreedom] -= alpha;
    if( theFreedom == 0 ) coord = coords[0]; 
        
  }// end multiple degree freedom
     
} // end function


void Matrix::recenter(int theFreedom, double shiftAccuracy)
{
#ifndef OPTIMIZE_QSD
  if( (theFreedom >= nFreedoms) || (theFreedom < 0) )
    error("Illegal degree of freedom in Matrix::recenter!");


  if( nFreedoms == 1 ) { // single degree of freedom
    if( (myType != FIELD) && (myType != PARTICLE) )
       error("Only types FIELD and PARTICLE can be recentered!");
  }
  else { // multiple degrees of freedom
    if( (freedomTypes[theFreedom] != FIELD) &&
        (freedomTypes[theFreedom] != PARTICLE) )
      error("Only types FIELD and PARTICLE can be recentered!");
  }
#endif
  if( nFreedoms == 1 ) {
    if (myType == FIELD)
    {
       
       fieldRecenter(theFreedom,shiftAccuracy);
     }
    if (myType == PARTICLE) particleRecenter(theFreedom,shiftAccuracy);
  }
  else {
    if (freedomTypes[theFreedom] == FIELD)
      fieldRecenter(theFreedom,shiftAccuracy);
    if (freedomTypes[theFreedom] == PARTICLE)
      particleRecenter(theFreedom,shiftAccuracy);
  }
}

void Matrix::fieldRecenter(int theFreedom, double shiftAccuracy)
{
  #ifndef OPTIMIZE_QSD
  if( (theFreedom >= nFreedoms) || (theFreedom < 0) )
    error("Illegal degree of freedom in Matrix::recenter!");
  if( nFreedoms == 1 )
  
   {
    if (myType != FIELD) error("Only type FIELD can be recentered!");
  }
  else {
    if (freedomTypes[theFreedom] != FIELD)
      error("Only type FIELD can be recentered!");
  }
#endif
  int i,j,k,l,numcol;
  
  //Matrix dummy(this->totalDim, FIELD);
  Matrix dummy=(*this);
  
    
  Complex magnitude, alpha;
  magnitude = this->Tr();
  
  if( magnitude == 0 ) error("Can't recenter zero state!");
  if( nFreedoms == 1 ) {		// 1 freedom case
    if( mySize > intSqrtMax ) 
      setIntSqrtMax(mySize*mySize);
    for(numcol = 0; numcol < mySize; numcol++)
    {
      
        dummy.myPointer = dummy.data + numcol*maxSize;
        
        for(k=1; k < mySize; k++)	// apply local annihilation op.
        { dummy.myPointer[(k-1)] =
            intSqrtList[k]*dummy.myPointer[k];
        }
        
      dummy.myPointer[mySize-1] = 0;
    }
  }
  else {				// multi-freedom case
    mySize = sizes[theFreedom];
    nSkip = nSkips[theFreedom];
    if( mySize > intSqrtMax ) 
      setIntSqrtMax(mySize);
    
    for (numcol = 0; numcol < maxSize; numcol++)
    { 
      for (i = 0; i < nSkip; i++)		// apply local annihilation op
      {
         for (j = 0; j < maxSize; j+=partDims[theFreedom]) 	// loop over outer indices
         {
           dummy.myPointer = dummy.data + i + j + numcol*maxSize;		
           for (k=1,l=nSkip; k<mySize; k++,l+=nSkip)
             dummy.myPointer[(l-nSkip)] = 
                 intSqrtList[k]*dummy.myPointer[l];
      
           dummy.myPointer[(mySize-1)*nSkip] = 0;
         }
       }
     }    
    mySize = maxSize;
    nSkip = 1;
    dummy.myPointer = dummy.data;
  }
  
  
  alpha = - (dummy.Tr()/magnitude);	
  
    if( abs(alpha) > shiftAccuracy )
      moveCoords2(alpha,theFreedom);	// shift basis
    //moveCoords(alpha,theFreedom, shiftAccuracy);	// shift basis
   
}


void Matrix::particleRecenter(int theFreedom, double shiftAccuracy)
{}

void Matrix::moveToCoords(const Complex& alpha, int theFreedom, double shiftAccuracy)
{
#ifndef OPTIMIZE_QSD
  if( (theFreedom < 0) || (theFreedom >= nFreedoms) )
    error("Illegal degree of freedom passed in State::moveToCoords!");
  if( nFreedoms == 1 ) {
    if( myType != FIELD )
      error("Can only change basis of FIELD freedoms in State::moveToCoords!");
  }
  else {
    if( freedomTypes[theFreedom] != FIELD )
      error("Can only change basis of FIELD freedoms in State::moveToCoords!");
  }
#endif
  Complex displacement;
  displacement = getCoords(theFreedom) - alpha;
  moveCoords(displacement,theFreedom,shiftAccuracy);
}


void Matrix::centerOn(Matrix& rho, double shiftAccuracy)
{
#ifndef OPTIMIZE_QSD
  if( nFreedoms != rho.nFreedoms )
    error("Mismatched number of degrees of freedom in State::centerOn!");
#endif
  Complex alpha;
  if( nFreedoms == 1 ) {
    if( (myType != FIELD) || (rho.myType != FIELD) )
      error("Can't change basis of non-FIELD type in State::centerOn!");
    alpha = centerCoords() - rho.centerCoords();
    moveCoords(alpha,0,shiftAccuracy);
  }
  else {
    for (int i=0; i<nFreedoms; i++) {
      if( freedomTypes[i] != rho.freedomTypes[i] )
        error("Mismatched types in State::centerOn!");
      if( freedomTypes[i] == FIELD ) {
        alpha = getCoords(i) - rho.getCoords(i);
        moveCoords(alpha,i);
      }
    }
  }
}


void Matrix::adjustCutoff(int theFreedom, double epsilon, int padSize)
//
// This member function adjusts the amount of memory actually being
// used by a Matrix.  As very frequently only the lowest few states have
// significant probability, it is more efficient not to loop over the
// top states at all.  This routine eliminates the top N states (which
// contain total probability less than epsilon) with a pad of size padSize
// for safety.
//
// We could consider different pad sizes for row and column, but for now
// we'll assume they are equal.  We will be dealing mostly with square 
// matrices.
{
  
#ifndef OPTIMIZE_QSD
  if( (theFreedom < 0) || (theFreedom >= nFreedoms) )
    error("Illegal freedom passed to Matrix::checkBounds!");
  if( (nFreedoms == 1) && (padSize > totalDim) )
    error("Too large a padSize requested in Matrix::adjustCutoff!");
  if( (nFreedoms > 1) && (padSize > nDims[theFreedom]) )
    error("Too large a padSize requested in Matrix::adjustCutoff!");
  if( padSize < 1 ) error("padSize must be positive in State::adjustCutoff!");
  if( (epsilon < 0.0) || (epsilon > 1.0) )
    error("Illegal value of epsilon in Matrix::adjustCutoff!");
  
    
  if( (nFreedoms == 1) && (myType != FIELD) && (myType != PARTICLE) )
    error("Can only adjust cutoff size for FIELD and PARTICLE! In Matrix::adjustCutoff");
  if( (nFreedoms > 1) && (freedomTypes[theFreedom] != FIELD) &&
      (freedomTypes[theFreedom] != PARTICLE) )
    error("Can only adjust cutoff size for FIELD and PARTICLE! In Matrix::adjustCutoff");
    
    
#endif
  Complex sum = 0.0;
  int i,j,l,m,n, row, col;
  
  this->normalize();				// make sure total prob. is 1
  
  if( nFreedoms == 1) {			// 1 freedom case
    for(i=maxSize-1; sum.re < epsilon;i--)
    {
       sum += data[i + maxSize*i];		// add up probability
    }
    
    i += (padSize + 2);
 
    if( i > totalDim)  {		// enough states allocated?
      cerr << "Warning!  Significant probability in top states!" << endl;
      i = totalDim;
    }
    if( i > maxSize )				// stretch # of states;
    {
      
      for(m = maxSize*i; m < i*i; m++)
      {
        data[m] = 0;
      }
      for(col = maxSize-1; col >= 0; col--)
      {  
        for(row = maxSize-1; row >= 0; row--)
        {
          data[row+col*i] = data[row+col*maxSize];
        } 
        for(row = maxSize; row < i; row++)
            data[row+col*i] = 0;
      }
    }     
     
    if (i < maxSize)
    {
      for(col = 0; col < i; col++)
      {
        for(row = 0; row < i; row++)
        { 
          data[row+col*i] = data[row+col*maxSize];
        }
      }
    } 
             
    maxSize = i;
    mySize = maxSize;
  }
  else {				// multi-freedom case
    
    int theSkip = nSkips[theFreedom];
    int theBound = (sizes[theFreedom] - 1)*theSkip;
    int theDim = partDims[theFreedom];
    int k;
// sum over other indices
    for(k=sizes[theFreedom],l=theBound; sum.re < epsilon; k--,l-=theSkip)
    for(i=0; i<theSkip; i++)
    for(j=0; j<maxSize; j+=theDim) {
      sum += data[(i+j+l)+(maxSize*(i+j+l))];		// add up probability
    }
    k += (padSize + 1);
    if( k > nDims[theFreedom] ) {	// enough states allocated?
      cerr << "Warning!  Significant probabilty in top states!" << endl;
      cerr << "k = " << k << endl;
      k = nDims[theFreedom];
    }
    if( k > sizes[theFreedom] ) stretchFreedom(theFreedom,k);
    if( k < sizes[theFreedom] ) shrinkFreedom(theFreedom,k);
  }

 
}


void Matrix::fullSize()
{
  int i,j;
  if(nFreedoms == 1)
  {
    for(i = maxSize; i < totalDim; i++)
      for(j = maxSize; j < totalDim; j++)
        data[i + j*totalDim] = 0.0; 
    
    maxSize = totalDim;
    mySize = totalDim;
  }
  else
  {
    for(i = 0; i < nFreedoms; i++)
      if(sizes[i] < nDims[i])
        stretchFreedom(i,nDims[i]);
  }
  
}


///////////////////////////////////////////////////////////////////

void Matrix::error(const char* message) const	// error handling
{
  cerr << message << "\n";
  exit(1);
}

int Matrix::getmySize()
{
  return mySize;
}

FreedomType Matrix::getmyType()
{
  return myType;
}

int Matrix::getmaxSize()
{
  return maxSize;
}

Complex& Matrix::getdata(int i, int j)
{
  return data[i + totalDim*j];
}

void Matrix::setdata(int i, int j, Complex z)
{
  if (i < 0 || j < 0)
    error("Matrix::setdata array index out of bounds");
  data[i + totalDim*j] = z;
}

void Matrix::printMatrix()
{
  int i,j;
  cout << "Printing ... " << endl;
  for(i = 0; i < maxSize; i++)
    for(j = 0; j < maxSize; j++)
      cout << "[" << i << "][" << j << "]=" << getdata(i,j) << endl;
  
     
}      
      
     
     
      
      
      
