#include <iostream.h>
#include <math.h>
#include <stdlib.h>

#include "Matrix.h"

//#include "PrimOp.h" // not sure if we need this in Matrix

static double smallSize = 1.0e-8;
static int intSqrtMax = 0;


Matrix::Matrix()
{
  nFreedoms = 0;
  totalDim = 0;
  maxSize = 0;
  nDims = 0;
  sizes = 0;
  nSkips = 0;
  partDims = 0;
  freedomTypes = 0;
  coords = 0;
  data = 0;
  myPointer = 0;
  nSkip = 0;
  myType = ALL;
  mySize = 0;
  coord = 0;
}

Matrix::Matrix(int n, FreedomType form)
{
  #ifndef OPTIMIZE_QSD
  if (n<2) error("Invalid dimension size in Matrix constructor!");
  if( (form==SPIN) && (n != 2) ) error("Spins must have dimension 2!");
  #endif
  	
  nFreedoms = 1;
  totalDim = maxSize = mySize = n;
  nSkip = totalDim;
  coord = 0;
  myType = form;
  data = new Complex[totalDim*totalDim];
  myPointer = data;
  nDims = 0;
  sizes = 0;
  nSkips = 0;
  partDims = 0;
  freedomTypes = 0;
  coords = 0;
  
  for (int i = 0; i < totalDim; i++)
  {
    for (int j = 0; j < totalDim; j++)
    {
      (*this)[i][j] = 0;
    }
  }
	
  data[0] = 1.0;// ground state
}
		

Matrix::Matrix(Matrix& rho)
{
  nFreedoms = 0;
  totalDim = 0;
  copy(rho);
}

Matrix::Matrix(int n, Complex alpha, FreedomType form)     // Constructor for
{							  																					// coherent state
#ifndef OPTIMIZE_QSD
  if (n<2) error("Invalid dimension size in State constructor!");
  if( form != FIELD )
    error("Constructor only valid for FIELD type.");
#endif
  data = 0;
  coherent(n,alpha);		// call coherent state function
}

void Matrix::coherent(int n, Complex alpha)	// Create a coherent state
{
#ifndef OPTIMIZE_QSD
  if( n < 2 )
    error("Invalid state size in State::coherent!");	 // exit on error
  if( norm(alpha) > n )
    error("Coherent state too big for storage in State::coherent!");
#endif
  Complex z1;
  Complex z2;
  free();				// recycle memory
  totalDim = maxSize = mySize = n;	// set up state
  nFreedoms = 1;
  nSkip = totalDim;
  coord = 0;
  myType = FIELD;
  data = new Complex[totalDim*totalDim];		// allocate memory
  myPointer = data;
  nDims = 0;
  sizes = 0;
  nSkips = 0;
  partDims = 0;
  freedomTypes = 0;
  coords = 0;
  z1 = exp( - 0.5*norm(alpha));		// normalization factor
  for (int i = 0; i < totalDim; i++) 
  {
    z2 = 1;
    for (int j = 0; j < totalDim; j++)
    {
      (*this)[i][j] = z1*z2;			// coherent state elements
      z2 *= alpha/sqrt(j+1);
    }
    z1 *= alpha/sqrt(i+1);
  }
}


Matrix::Matrix(int n, Matrix* elements, FreedomType form) 
//
// Constructor for general 1 degree of freedom state
{
#ifndef OPTIMIZE_QSD
  if (n<2) error("Invalid dimension size in State constuctor!");
  if( (form==SPIN) && (n != 2) ) error("Spins must have dimension 2!");
#endif
  nFreedoms = 1;
  totalDim = maxSize = mySize = n;
  nSkip = totalDim;
  coord = 0;
  myType = form;
  data = new Complex[totalDim*totalDim];
  myPointer = data;
  nDims = 0;
  sizes = 0;
  nSkips = 0;
  partDims = 0;
  freedomTypes = 0;
  coords = 0;
  for (int i=0; i < totalDim; i++) 
	{
		for (int j = 0; j < totalDim; j++)
		{
			(*this)[i][j] = (*elements)[i][j]; 
		}
	}
}

Matrix::Matrix(int n, int* dimensions, FreedomType* forms) // Constructor for
{							 // multidimensional
  int i,j;						 // ground state
#ifndef OPTIMIZE_QSD
  if (n<1) error("Illegal number of freedoms in State constructor!");
  if (n=1) error("Do not call this constructor for 1-freedom states!");
#endif
  nFreedoms = n;
  nDims = new int[n];			// allocate memory
  sizes = new int[n];
  nSkips = new int[n];
  partDims = new int[n];
  freedomTypes = new FreedomType[n];
  coords = new Complex[n];
  totalDim = 1;
  for (i=0; i<n; i++) {
#ifndef OPTIMIZE_QSD
    if(dimensions[i] < 1) error("Illegal dimension in State constructor!");
    if( (forms[i] == SPIN) && (dimensions[i] != 2) )
      error("Spin states must have 2 dimensions!");
#endif
    nSkips[i] = totalDim;
    totalDim *= (nDims[i] = dimensions[i]);
    sizes[i] = nDims[i];
    partDims[i] = totalDim;
    freedomTypes[i] = forms[i];
    coords[i] = 0;
  }
  maxSize = totalDim;
  data = new Complex[totalDim*totalDim];		// allocate memory
  myPointer = data;
  mySize = totalDim;
  nSkip = 1;
  coord = 0;
  myType = freedomTypes[0];
  data[0] = 1;
  for (i=1; i<totalDim; i++) 
    for (j=1; j < totalDim; j++)
        (*this)[i][j] = 0;

}
		
Matrix::Matrix(int n, int nstate, FreedomType form)	// Constructor for 
{							// Fock state
#ifndef OPTIMIZE_QSD
  if (n<2) error("Invalid dimension size in State constructor!");
  if( form != FIELD )
    error("Constructor only valid for FIELD type.");
#endif
  data = 0;
  fock(n,nstate);		// call Fock state function
}


void Matrix::fock(int n, int nstate)	// Create a Fock State
{
#ifndef OPTIMIZE_QSD
  if( (nstate >= n) || (n < 2) || (nstate < 0) )	// check dimension
    error("Illegal arguments to Matrix::fock!");		// exit on error
#endif
  free();				// recycle memory
  totalDim = maxSize = mySize = n;	// set up state
  nFreedoms = 1;
  nSkip = totalDim;
  coord = 0;
  myType = FIELD;
  data = new Complex[totalDim*totalDim];		// allocate memory
  myPointer = data;
  nDims = 0;
  sizes = 0;
  nSkips = 0;
  partDims = 0;
  freedomTypes = 0;
  coords = 0;
  for (int i = 0; i < totalDim; i++) 
  {
    for (int j = 0; j < totalDim; j++)
    {
      (*this)[i][j] = 0;
    }
  }
	
  (*this)[nstate][nstate] = 1;					// Fock state
}


Matrix::Matrix(int n, Complex alpha, double width, double dx, FreedomType form)
{					// Constructor for Gaussian state
#ifndef OPTIMIZE_QSD
  if (n<2) error("Invalid dimension size in State constructor!");
  if( form != PARTICLE )
    error("Constructor only valid for PARTICLE type.");
#endif
  data = 0;
  gaussian(n,alpha,width,dx);		// call Gaussian state function
}


void Matrix::gaussian(int n, Complex alpha, double width, double dx)
{					// Create a Gaussian state
// The PARTICLE type is a position representation.  It stores the center
// of coordinates in the real part of coord.  In the imaginary part, it
// stores the lattice spacing dx.  If imag(coord) == 0, then it assumes
// that dx=1.  The amplitudes are stored in the data array with the
// negative locations in the odd elements and the nonnegative in the
// even elements.
#ifndef OPTIMIZE_QSD
  if( n < 2 )
    error("Invalid state size in Matrix::gaussian!");	 // exit on error
  if( dx < 0 )
    error("Invalid lattice spacing in Matrix::gaussian!"); // exit on error
  if( (dx != 0) && (width/dx > n) )
    error("Gaussian state too big for storage in Matrix::gaussian!");
  if( (dx == 0) && (width > n) )
    error("Gaussian state too big for storage in Matrix::gaussian!");
#endif
  Complex p(0.0,0.5*alpha.im*dx);	// center of Gaussian in momentum
  double coeff = 0.25*dx*dx/(width*width);	// coefficient in exponent
  Complex phase(0.0,alpha.im*alpha.re);
  free();				// recycle memory
  totalDim = maxSize = mySize = n;	// set up state
  nFreedoms = 1;
  nSkip = totalDim;
  Complex z(alpha.re,dx);
  coord = z;				// center of Gaussian, lattice space
  myType = PARTICLE;
  data = new Complex[totalDim*totalDim];		// allocate memory
  myPointer = data;
  nDims = 0;
  sizes = 0;
  nSkips = 0;
  partDims = 0;
  freedomTypes = 0;
  coords = 0;
  int i, j;
  
  // even-even
  for(i = 0; i < totalDim; i += 2)
  {
    for(j = 0; j < totalDim; j += 2)
    {
      (*this)[i][j] = exp(-coeff*(i*i + j*j) + p*(i - j) + phase);
    }
  }
  
  // even-odd
  
  for(i = 0; i < totalDim; i +=2)
  {
    for(j = 2; j < totalDim; j += 2)
    {
      (*this)[i][j-1] = exp(-coeff*(i*i + j*j) + p*(i - j) + phase);
    }
  }
  
  // odd-even
  
  for(i = 2; i < totalDim; i += 2)
  {
    for(j = 0; j < totalDim; j += 2)
    {
      (*this)[i-1][j] = exp(-coeff*(i*i + j*j) - p*(i + j) + phase);
    }
  }
  
  // odd-odd
  
  for(i = 2; i < totalDim; i += 2)
  {
    for(j = 2; j < totalDim; j += 2)
    {
      (*this)[i-1][j-1] = exp(-coeff*(i*i + j*j) - p*(i + j) + phase);
    }
  }
  normalize();
}


Matrix::~Matrix()
{
  free();
}
	
void Matrix::free()
{
  if(data != 0)
  {
#ifndef NON_GNU_DELETE
    delete[] data;
    if(nFreedoms > 1)
    {
      delete[] nDims;
      delete[] sizes;
      delete[] nSkips;
      delete[] partDims;
      delete[] freedomTypes;
      delete[] coords;
    }
#else // ---- NON GNU CODE ----
    delete [totalDim] data;		// NON GNU recycle storage
    if( nFreedoms > 1 ) {
      delete [nFreedoms] nDims;		// NON GNU recycle parameter arrays
      delete [nFreedoms] sizes;
      delete [nFreedoms] nSkips;
      delete [nFreedoms] partDims;
      delete [nFreedoms] freedomTypes;
      delete [nFreedoms] coords;
    }
#endif // ---- NON GNU CODE ----
    nFreedoms = 0;			// zero variables
    totalDim = 0;
    maxSize = 0;
    myType = ALL;
    mySize = 0;
    nSkip = 0;
    coord = 0;
    myPointer = 0;			// reset pointers to zero
    data = 0;
    nDims = 0;
    sizes = 0;
    nSkips = 0;
    coords = 0;
    partDims = 0;
    freedomTypes = 0;
  }
}
	

void Matrix::normalize()
{
  int i,j;
  /*double sum=0.0;
  double epsLimit;
  epsLimit = smallSize/maxSize;
  */
  
      
  Complex trace = this->Tr();
  if (trace == 0)
    error("Division by zero trace not allowed Matrix::normalize()");
       
  Matrix* temp = new Matrix(mySize);
      
  for (i = 0; i < mySize; i++)
  {
    for (j = 0; j < mySize; j++)
    {
      (*temp)[i][j] = (*this)[i][j]/trace;
    }
  }
  
  for (i = 0; i < mySize; i++)
  {
    for (j = 0; j < mySize; j++)
    {
      (*this)[i][j] = (*temp)[i][j];
    }
  }
        
        
}

void Matrix::xerox(Matrix& rho)
// Unlike copy, xerox makes a copy of an existing matrix including only
// the memory which is actually being used, rather than all the memory
// allocated.  This is useful as a more efficient way of creating
// temporary matrices than using the normal copy-constructor.
//
// Note that xerox should NOT be used to assign to a matrix which already
// has memory allocated.  This results in inefficient memory handling, and
// might cause errors.
{
  int i,j;
#ifndef OPTIMIZE_QSD
  if( data != 0 )
    error("Matrix::xerox should not be used on a matrix with allocated memory!");
#endif
  nFreedoms = rho.nFreedoms;
  if( nFreedoms > 1 ) {
    nDims = new int[nFreedoms];
    sizes = new int[nFreedoms];
    nSkips = new int[nFreedoms];
    partDims = new int[nFreedoms];
    freedomTypes = new FreedomType[nFreedoms];
    coords = new Complex[nFreedoms];
  }
  totalDim = rho.maxSize;
  data = new Complex[totalDim*totalDim];
  maxSize = rho.maxSize;			// copy member data
  nSkip = rho.nSkip;
  coord = rho.coord;
  mySize = rho.mySize;
  myType = rho.myType;
  myPointer = data;
  for (i=0; i < maxSize; i++)
    for (j = 0; j < maxSize; j++)
      (*this)[i][j] = rho[i][j];
  
  if( nFreedoms > 1 ) {
    for (i=0; i<nFreedoms; i++) {
      nDims[i] = rho.sizes[i];
      sizes[i] = rho.sizes[i];
      nSkips[i] = rho.nSkips[i];
      partDims[i] = rho.partDims[i];
      freedomTypes[i] = rho.freedomTypes[i];
      coords[i] = rho.coords[i];
    }
  }
}

Matrix& Matrix::operator=(Matrix& rho)
{
  if (this != &rho)// otherwise rho = rho would lead to trouble
  {
    if((totalDim != rho.totalDim) || (nFreedoms != rho.nFreedoms))
    {
      free();//recycle any memory
    }
		
   copy(rho);//copy rho into b
  }
	
  return *this;
}

void Matrix::copy(Matrix& rho)
{
  int i,j;
  if( nFreedoms != rho.nFreedoms)
  {
    nFreedoms = rho.nFreedoms;
    if( nFreedoms > 1)
    {
      nDims = new int[nFreedoms];
      sizes = new int[nFreedoms];
      nSkips = new int[nFreedoms];
      partDims = new int[nFreedoms];
      freedomTypes = new FreedomType[nFreedoms];
      coords = new Complex[nFreedoms];
    }
  }
	
  if( totalDim != rho.totalDim ) 
  {	// if no memory allocated, allocate it
    totalDim = rho.totalDim;
    data = new Complex[totalDim*totalDim];
  }
  maxSize = rho.maxSize;			// copy member data
  nSkips = rho.nSkips;
  coord = rho.coord;
  mySize = rho.mySize;
  myType = rho.myType;
  myPointer = data;
  for (i = 0; i < maxSize; i++)
  {
    for (j = 0; j < maxSize; j++)
    {
      (*this)[i][j] = rho[i][j];
    
    }
  }
  if( nFreedoms > 1 ) 
  {
   for (i = 0; i < nFreedoms; i++) 
   {
      nDims[i] = rho.nDims[i];
      sizes[i] = rho.sizes[i];
      nSkips[i] = rho.nSkips[i];
      partDims[i] = rho.partDims[i];
      freedomTypes[i] = rho.freedomTypes[i];
      coords[i] = rho.coords[i];
    }
  }
}
	
Matrix& Matrix::operator=(int n)
{
  if(n != 0)
    error("Cannot assign a Matrix a value other than zero!");
  for(int i = 0; i < maxSize; i++)
  {
    for(int j = 0; j < maxSize; j++)
    {
      (*this)[i][j] = 0;
    }
  }
	
  return *this;
}			 


Complex* Matrix::operator[](int i)
{
  
  return myPointer+totalDim*i;
}
	
Matrix& Matrix::operator*=(Matrix& rho)
{
  int i,j,k;
  if(mySize != rho.mySize)
    error("Incompatible matrix sizes in Matrix::operator*!");
  if(myType != rho.myType)
    error("Incompatible matrix types in Matrix::operator*!");
	
  Matrix* temp = new Matrix(mySize);
	
  for(i = 0; i < mySize; i++)
  {
    for(j = 0; j < mySize; j++)
    {
      Complex sum = 0.0;
			
      for(k = 0; k < mySize; k++)
      {
        sum += (rho[i][k])*(rho[k][j]);
      }
			
      (*temp)[i][j] = sum;
    }
  }
	
  for(i = 0; i < mySize; i++)
  {
    for(j = 0; j < mySize; j++)
    {
      (*this)[i][j] = (*this)[i][j]*(*temp)[i][j];
    }
  }	
  return *this;
}

Matrix& Matrix::operator*=(const Complex& z)
{
  for (int i = 0; i < mySize; i++)
  {
    for (int j = 0; j < mySize; j++)
    {
      (*this)[i][j] = (*this)[i][j]*z;
    }
  }
  
  return *this;
}

Matrix& Matrix::operator*=(const double x)
{
  for (int i = 0; i < mySize; i++)
  {
    for (int j = 0; j < mySize; j++)
    {
      (*this)[i][j] = (*this)[i][j]*x;
    }
  }
  
  return *this;
}

Matrix& Matrix::operator*=(ImaginaryUnit im)
{
  int i,j;
  if(im == IM)
  {
    for(i = 0; i < mySize; i++)
    {
      for(j = 0; j < mySize; j++)
      {
        (*this)[i][j].timesI();
      }
    }
  }
  if (im == M_IM)
  {
    for(i = 0; i < mySize; i++)
    {
      for(j = 0; j < mySize; j++)
      {
        (*this)[i][j].timesMinusI();
      }
    }
  }

  return *this;
}


Matrix& Matrix::operator+=(Matrix& rho)
{
  if(mySize != rho.mySize)
    error("Incompatible matrix sizes in Matrix::operator*!");
  if(myType != rho.myType)
    error("Incompatible matrix types in Matrix::operator*!");
  for(int i = 0; i < mySize; i++)
  {
    for(int j = 0; j < mySize; j++)
    {
      (*this)[i][j] = rho[i][j] + (*this)[i][j];
    }	  
  }
	
  return *this;
}


Matrix& Matrix::operator-=(Matrix& rho)
{
  if(mySize != rho.mySize)
    error("Incompatible matrix sizes in Matrix::operator*!");
  if(myType != rho.myType)
    error("Incompatible matrix types in Matrix::operator*!");
  for(int i = 0; i < mySize; i++)
  {
    for(int j = 0; j < mySize; j++)
    {
      (*this)[i][j] = rho[i][j] - (*this)[i][j];
    }	  
  }
	
  return *this;
}
	
//Trace of a matrix
Complex Matrix::Tr()
{
  int i;
  Complex sum = 0.0;
	
  for(i = 0; i < mySize; i++)
  {
    sum += (*this)[i][i];
  }
	
  return sum;
}
	

//Transpose of a matrix
Matrix& Matrix::transpose()
{
  int i,j;
  Matrix* temp = new Matrix(mySize);
  for(i = 0; i < mySize; i++)
  {
    for(j = 0; j < mySize; j++)
    {
      (*temp)[i][j] = (*this)[j][i];
    } 
  }
	
  for(i = 0; i < mySize; i++)
  {
    for(j = 0;j < mySize; j++)
    {
      (*this)[i][j] = (*temp)[i][j];
    }
  } 
  return *this;
}

//Conjugate of a matrix
Matrix& Matrix::conjugate()
{
  int i,j;
  Matrix* temp = new Matrix(mySize);
  for(i = 0; i < mySize; i++)
  {
    for(j = 0; j < mySize; j++)
    {
      (*temp)[i][j] = conj((*this)[i][j]);
    }
  }
  
  for(i = 0; i < mySize; i++)
  {
    for(j = 0; j < mySize; j++)
    {
      (*this)[i][j] = (*temp)[i][j];
    }
  }
  return *this;
}

//Hermitian Conjugate
Matrix& Matrix::hc()
{
  int i,j;
  Matrix* tempconj = new Matrix(mySize);
  Matrix* temptrans = new Matrix(mySize);
  for(i = 0; i < mySize; i++)
  {
    for(j = 0; j < mySize; j++)
    {
      (*tempconj)[i][j] = conj((*this)[i][j]);
    
    }
  }
  
  for(i = 0; i < mySize; i++)
  {
    for(j = 0; j < mySize; j++)
    {
      (*temptrans)[i][j] =(*tempconj)[j][i];
      
    }
  }
  
  for(i = 0; i < mySize; i++)
  {
    for(j = 0; j < mySize; j++)
    {
      
      (*this)[i][j] = (*temptrans)[i][j];
    }
  }
  
  return *this;  
}	
					
		
//Output a matrix to an output stream
ostream& operator<<( ostream& s, Matrix& rho)
{
  int i, j;
  s << rho.nFreedoms << endl;
  s << rho.totalDim << endl;
  s << rho.maxSize << endl;
  s << rho.mySize << endl;
  s << rho.nSkip << endl;
  s << rho.coord << endl;
  s << rho.myType << endl;
  if(rho.nFreedoms > 1)
  {
    for (i = 0; i < rho.nFreedoms; i++)
    {
      s << rho.nDims[i] << endl;
      s << rho.sizes[i] << endl;
      s << rho.nSkips[i] << endl;
      s << rho.partDims[i] << endl;
      s << rho.freedomTypes[i] << endl;
      s << rho.coords[i] << endl;
    }
    for (i = 0; i < rho.maxSize; i++)
    {
      for (j = 0; j < rho.maxSize; j++)
      {
        s << rho[i][j] << endl;
      }
    }
  }
	
  return s;
} 

istream& operator>>( istream& s, Matrix& rho )
//
// Input a State from an input stream
//
{
  int i,j;

  rho.free();			// free up State
  s >> rho.nFreedoms;
  s >> rho.totalDim;
  s >> rho.maxSize;
  s >> rho.mySize;
  s >> rho.nSkip;
  s >> rho.coord;
  s >> rho.myType;
  if( rho.totalDim > 0 ) {
    rho.data = new Complex[rho.totalDim];
    rho.myPointer = rho.data;
  }
  if( rho.nFreedoms > 1 ) {
    rho.nDims = new int[rho.nFreedoms];
    rho.sizes = new int[rho.nFreedoms];
    rho.nSkips = new int[rho.nFreedoms];
    rho.partDims = new int[rho.nFreedoms];
    rho.freedomTypes = new FreedomType[rho.nFreedoms];
    rho.coords = new Complex[rho.nFreedoms];
    for(i=0; i<rho.nFreedoms; i++) 
		{
      	s >> rho.nDims[i];
      	s >> rho.sizes[i];
      	s >> rho.nSkips[i];
      	s >> rho.partDims[i];
      	s >> rho.freedomTypes[i];
      	s >> rho.coords[i];
    }
  }
  
  for (i = 0; i < rho.maxSize; i++)
  {
    for(j = 0; j < rho.maxSize; j++)
    { 	
      s >> rho[i][j];
    }
  }
  
  return s;
}
	

Matrix operator*(Matrix& rho, Matrix& phi)
{
	Matrix result(rho);
	return (result *= phi);
}

		
Matrix operator*(const Complex& z, Matrix& rho)
{
  Matrix result(rho);
  return (result *= z);
}

Matrix operator*(Matrix& rho, const Complex& z)
{
  Matrix result(rho);
  return (result *= z);
}

Matrix operator*(const double x, Matrix& rho)
{
  Matrix result(rho);
  return (result *= x);
}

Matrix operator*(Matrix& rho, const double x)
{
  Matrix result(rho);
  return (result *= x);
}

Matrix operator*(ImaginaryUnit i, Matrix& rho)
{
  Matrix result(rho);
  return (result *= i);
}

Matrix operator*(Matrix& rho, ImaginaryUnit i)
{
  Matrix result(rho);
  return (result *= i);
}

Matrix& operator+(Matrix& rho, Matrix& phi)
{
  Matrix& result(rho);
  return (result += phi);
}
	
Matrix& operator-(Matrix& rho, Matrix& phi)
{
  Matrix& result(rho);
  return (result -= phi);
} 

Matrix& operator+(Matrix& rho)
{
  Matrix& result(rho);
  return result;
}

Matrix& operator-(Matrix& rho)
{
  Matrix& result(rho);
   
  int getmaxSize = result.getmaxSize();
  for (int i = 0; i < getmaxSize; i++)
  {
    for (int j = 0; j < getmaxSize; j++)
    {
      result[i][j] = -result[i][j];
    }
  }
	
  return result;
}

int Matrix::size()
{
  return mySize;
}

int Matrix::getSize(int theFreedom)
{
  #ifndef OPTIMIZE_QSD
  if( (theFreedom < 0) || (theFreedom >= nFreedoms) )
    error("Illegal freedom passed to State::getSize!");
#endif
  return sizes[theFreedom];
}

void Matrix::diagnostic()	// for debugging
{
  int i;
  cout << "Number of Freedoms " << nFreedoms << ".\n";
  cout << "Total Dimensions " << totalDim << ".\n";
  cout << "MaxSize " << maxSize << ".\n";
  cout << "My Type " << myType << ".\n";
  cout << "My Size " << mySize << ".\n";
  cout << "nSkip " << nSkip << ".\n";
  cout << "coord " << coord << ".\n";
  cout << "data pointer " << data << ".\n";
  cout << "My Pointer " << myPointer << ".\n";
  cout << "Dimensions:\n";
  if (nFreedoms > 1 ) {
    for (i=0; i<nFreedoms; i++)
      cout << "  " << nDims[i] << ".\n";
    cout << "Sizes:\n";
    for (i=0; i<nFreedoms; i++)
      cout << "  " << sizes[i] << ".\n";
    cout << "Skips:\n";
    for (i=0; i<nFreedoms; i++)
      cout << "  " << nSkips[i] << ".\n";
    cout << "Partial dimensions:\n";
    for (i=0; i<nFreedoms; i++)
      cout << "  " << partDims[i] << ".\n";
    cout << "Centers of coordinates:\n";
    for (i=0; i<nFreedoms; i++)
      cout << "  " << coords[i] << ".\n";
  }

}
	


void Matrix::error(const char* message) const	// error handling
{
  cerr << message << "\n";
  exit(1);
}

int Matrix::getmySize()
{
  return mySize;
}

FreedomType Matrix::getmyType()
{
  return myType;
}

int Matrix::getmaxSize()
{
  return maxSize;
}
