//   ParticleOp.cc -- Operators for particle in position representation.
//     
//   Copyright (C) 2002  Todd Brun and Ruediger Schack
//   
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//   
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//   
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//   ----------------------------------------------------------------------
//   If you improve the code or make additions to it, or if you have
//   comments or suggestions, please contact us:
//
//   Dr. Todd Brun			        Tel    +44 (0)171 775 3292
//   Department of Physics                      FAX    +44 (0)181 981 9465
//   Queen Mary and Westfield College           email  t.brun@qmw.ac.uk
//   Mile End Road, London E1 4NS, UK
//
//   Dr. Ruediger Schack                        Tel    +44 (0)1784 443097
//   Department of Mathematics                  FAX    +44 (0)1784 430766
//   Royal Holloway, University of London       email  r.schack@rhbnc.ac.uk
//   Egham, Surrey TW20 0EX, UK
/////////////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include <stdlib.h>

#include "ParticleOp.h"

void PositionOperator::applyTo(State& v, int hc, double)
{
#ifdef DEBUG_TRACE
  cout << "PositionOperator::applyTo entered." << endl;
#endif
  int i;
  Complex center = v.centerCoords();
  double x0 = center.real();
  double dx = 0.5*center.imag();
  if (dx == 0) dx = 0.5;
  int vSize = v.size();
  for( i=0; i<vSize; i+=2 )	// positive (even) positions
    v[i] *= (x0 + dx*i);
  for( i=2; i<vSize+1; i+=2 )	// negative (odd) positions
    v[i-1] *= (x0 - dx*i);
}

void MomentumOperator::applyTo(State& v, int hc, double)
{
#ifdef DEBUG_TRACE
  cout << "MomentumOperator::applyTo entered." << endl;
#endif
  int i;
  Complex center = v.centerCoords();
  double x0 = center.real();
  double dx = center.imag();
  if (dx == 0) dx = 1.0;
  Complex z(0.0,-0.5/dx);
  int vSize = v.size();
  int highestEven,highestOdd;
  if( (vSize%2) == 0) {
    highestEven = vSize-2;
    highestOdd = vSize-1;
  } else {
    highestEven = vSize-1;
    highestOdd = vSize-2;
  }
  Complex* temp;
  temp = new Complex[vSize];
  for( i=2; i<highestEven; i+=2 )	// positive (even) positions
    temp[i] = z*(v[i+2]-v[i-2]);
  for( i=3; i<highestOdd; i+=2 )	// negative (odd) positions
    temp[i] = z*(v[i-2]-v[i+2]);
  temp[0] = z*(v[2]-v[1]);
  temp[1] = z*(v[0]-v[3]);
  temp[highestEven] = -z*v[highestEven-2];
  temp[highestOdd] = z*v[highestOdd-2];
  for( i = 0; i<vSize; i++ ) v[i] = temp[i];
#ifndef NON_GNU_DELETE
  delete[] temp;
#else
  delete[vSize] temp;
#endif
}

void ShiftOperator::applyTo(State& v, int hc, double)
{
  int i;
  int vSize = v.size();
  int highestEven,highestOdd;
  if( (vSize%2) == 0) {
    highestEven = vSize-2;
    highestOdd = vSize-1;
  } else {
    highestEven = vSize-1;
    highestOdd = vSize-2;
  }
  switch( hc ) {
  case NO_HC:				// shift to the right
    for( i=highestEven; i>0; i-=2)	// truncate rightmost state
      v[i] = v[i-2];
    v[0] = v[1];			// -1 -> 0
    for( i=1; i<highestOdd; i+=2)
      v[i] = v[i+2];
    v[highestOdd] = 0;			// get 0 from left
    break;
  case HC:				// shift to the left
    for( i=highestOdd; i>1; i-=2)	// truncate leftmost state
      v[i] = v[i-2];
    v[1] = v[0];			// 0 -> -1
    for( i=0; i<highestEven; i+=2)
      v[i] = v[i+2];
    v[highestEven] = 0;			// get 0 from right
    break;
  default:
    error("Unknown option ShiftOperator::applyTo.");
  }
}

